import numpy as np
import isopy as isopy
from isopy import core

__all__ = ['remove_mass_fractionation', 'add_mass_fractionation',
           'calculate_mass_fractionation_factor', 'internal_normalisation',
           'mass_independent_correction',
           'remove_isobaric_interferences', 'find_isobaric_interferences',
           'make_ms_array', 'make_ms_beams', 'make_ms_sample', 'johnson_nyquist_noise',
           'rDelta', 'inverse_rDelta',
           'find_outliers', 'upper_limit', 'lower_limit']

import isopy.checks
import isopy.core


"""
Functions for isotope data reduction
"""

def johnson_nyquist_noise(voltage, resistor = 1E11, integration_time = 8.389, include_counting_statistics = True,
                          T = 309, R = 1E11, cpv = 6.25E7):
    """
    Calculate the Johnson-Nyquist noise and counting statistics for a given voltage.

    The Johnson-Nyquist noise (:math:`n_{jn}` is calculated as:

    .. math::
        n_{jn} = \\sqrt{ \\frac{4*k_{b}*T*r} {t} } * \\frac{R} {r}

    The counting statistics, or shot noise, (:math:`n_{cs}` is calculated as:

    .. math::
        n_{cs} = \\sqrt{ \\frac{1} {v * c_{V} * t}} * v

    The two are combined as:

    .. math::
        n_{all} = \\sqrt{ (n_{jn})^2 + (n_{cs})^2 }


    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Adapted from the equations in `Liu & Pearson (2014) Chemical Geology, 10, 301-311
    <https://doi.org/10.1016/j.chemgeo.2013.11.008>`_.

    Parameters
    ----------
    voltage : isotope_rarray, float, np.ndarray
        The measured voltages. :math:`v` in the equations above.
    resistor : isotope_rarray, float, np.ndarray, dict
        The resistor for the measurement. Default value is ``1E11``. :math:`r` in the equations above.
        If *resistor* is a dictionary *R* will be used for values not in the dictionary.
    integration_time : float, optional
        The integration time in seconds for a single measurement. Default value is ``8.389``.
        :math:`t` in the equations above.
    include_counting_statistics: bool, optional
        If ``True`` then the counting statistics are included in the returned value. Default value is ``True``.
    T : float, optional
        Amplifier housing temperature in kelvin. Default value is ``309``. :math:`T` in the equations above.
    R : float, optional
        *voltage* values are reported as volts for this resistor value. Default value is
        ``1E11``. :math:`R` in the equations above.
    cpv : float, optional
        Counts per volt per second. Default value is ``6.25E7``. :math:`C_{V}` in the equations above.

    Returns
    -------
    noise: np.float or np.ndarray or isotope_rarray
        The noise in V for the given voltage/set of voltages.

    Examples
    --------
    >>> isopy.tb.johnson_nyquist_noise(10) * 1000 #in millivolts
    0.13883808575503015
    >>> isopy.tb.johnson_nyquist_noise(10, 1E10) * 1000
    0.14528174343845432

    >>> array = isopy.tb.make_ms_array('pd')
    >>> array = array * (10 / array['106pd']) #10v on the largest isotope
    >>> isopy.tb.johnson_nyquist_noise(array) * 1000 #in millivolts
    (row) , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd
    None  , 0.03025 , 0.08932 , 0.12565 , 0.13884 , 0.13663 , 0.09156
    >>> resistors = dict(pd102=1E13, pd106=1E10) #1E11 is used for missing keys
    >>> isopy.tb.johnson_nyquist_noise(array, resistors) * 1000
    (row) , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd
    None  , 0.02672 , 0.08932 , 0.12565 , 0.14528 , 0.13663 , 0.09156
    """

    voltage = isopy.checks.check_type('voltage', voltage, isopy.core.IsopyArray, np.ndarray, np.float64, coerce=True,
                                      coerce_into=[isopy.core.IsopyArray, np.float64, np.array])
    resistor = isopy.checks.check_type('resistor', resistor, isopy.core.IsopyArray, np.ndarray, dict, coerce=True,
                                       coerce_into=[isopy.core.IsopyArray, np.float64, np.array])
    integration_time = isopy.checks.check_type('integration_time', integration_time, np.float64, coerce=True)
    include_counting_statistics = isopy.checks.check_type('include_counting_statistics', include_counting_statistics,
                                                          bool)
    if isinstance(resistor, (core.IsopyArray, dict)):
        resistor = core.ScalarDict(resistor, default_value=R)
        resistor = isopy.array({key: resistor.get(key) for key in voltage.keys})

    T = isopy.checks.check_type('T', T, np.float64, coerce=True)
    R = isopy.checks.check_type('R', R, np.float64, coerce=True)
    cpv = isopy.checks.check_type('cpv', cpv, np.float64, coerce=True)

    kB = np.float64(1.3806488E-023) # Boltsman constant

    t_noise = np.sqrt((4 * kB * T * resistor) / integration_time) * (R / resistor)

    if include_counting_statistics:
        c_stat = np.sqrt(1 / (voltage * cpv * integration_time)) * voltage
        return np.sqrt(c_stat ** 2 + t_noise ** 2)
    else:
        return t_noise

def make_ms_array(*args, mf_factor = None, isotope_fractions = None, isotope_masses=None, **kwargs):
    """
    Constructs an isotope array where the first created isotope contains the sum of all given
    isotopes with the same mass number. The construction order is first *args* and then *kwargs*.

    Parameters
    ----------
    args
        Each arg can be either a isotope array, a element key or an isotope key string.
        If arg is an isotope array an exception will be raised if two or more isotopes have the
        same mass number. If arg is an element key string then all isotopes of that element are
        added to the array. If arg is an isotope key string then only that isotope and isotopes
        with the same mass number as any other isotope in the array is added.
    mf_factor
        If given this mass fractionation factor is applied to all isotopes of an element
        prior to being added to the result.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.
    kwargs
        Each kwarg key must be either an element key string or an isotope key string. The isotope
        fractions added to the array based on the kwarg key are multiplied by the kwarg value.
        Otherwise behaves the same as for *args*.

    Returns
    -------
    result : IsopyArray
        The constructed isotope array.

    Examples
    --------
    >>> isopy.tb.make_ms_array('pd')
    (row) , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd
    None  , 0.01020 , 0.11140 , 0.22330 , 0.27330 , 0.26460 , 0.11720

    >>> isopy.tb.make_ms_array('pd', ru101=0.1)
    (row) , 101Ru   , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd
    None  , 0.01706 , 0.04175 , 0.13002 , 0.22330 , 0.27330 , 0.26460 , 0.11720

    >>> isopy.tb.make_ms_array('pd', ru101=0.1, ru99=0) #99Ru doesnt contribute anything to the array but gets a contribution from 101Ru.
    (row) , 99Ru    , 101Ru   , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd
    None  , 0.01276 , 0.01706 , 0.04175 , 0.13002 , 0.22330 , 0.27330 , 0.26460 , 0.11720

    >>> isopy.tb.make_ms_array('pd', 'cd')
    102Pd  , 104Pd  , 105Pd  , 106Pd   , 108Pd  , 110Pd   , 111Cd   , 112Cd   , 113Cd   , 114Cd   , 116Cd
    0.0102 , 0.1114 , 0.2233 , 0.28579 , 0.2735 , 0.24205 , 0.12804 , 0.24117 , 0.12225 , 0.28729 , 0.07501

    >>> isopy.tb.make_ms_array('pd', cd=1) #Same as example above
    (row) , 102Pd   , 104Pd   , 105Pd   , 106Pd   , 108Pd   , 110Pd   , 111Cd   , 112Cd   , 113Cd   , 114Cd   , 116Cd
    None  , 0.01020 , 0.11140 , 0.22330 , 0.28579 , 0.27350 , 0.24205 , 0.12804 , 0.24117 , 0.12225 , 0.28729 , 0.07501

    See Also
    --------
    make_ms_beams, make_ms_sample
    """

    isotope_fractions = isopy.checks.check_reference_value('isotope_fractions', isotope_fractions,
                                                           isopy.refval.isotope.fraction)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses,
                                                        isopy.refval.isotope.mass)

    input = []
    for key in args:
        if type(key) is tuple:
            input.append((k, 1) for k in key)
        elif isinstance(key, dict):
            kwargs.update(key)
        else:
            input.append((key, 1))

    for item in kwargs.items():
        input.append(item)


    out_keys = isopy.IsotopeKeyList()
    for key, val in input:
        if isinstance(key, core.IsopyArray):
            keys = key.keys()
        else:
            key = isopy.askeystring(key)
            if key.flavour in ['element', 'isotope',  'molecule']:
                keys = key.isotopes()
            else:
                raise ValueError(f'Unknown input: {key} input must be an element or isotope key string')

        keys = keys.filter(mz_neq = out_keys.mz())

        out_keys += tuple(dict(zip(keys.mz(), keys)).values())

    out_keys = out_keys.sorted()
    mz_keys = [key.mz() for key in out_keys]
    mzkey_map = dict(zip(mz_keys, out_keys))

    out = isopy.zeros(None, out_keys)
    for key, val in input:
        if isinstance(key, core.IsopyArray):
            array = key
        else:
            key = isopy.askeystring(key)
            keys = key.element().isotopes()
            keys = keys.filter(mz_eq=mz_keys)

            array = isopy.array(keys.fractions(isotope_fractions=isotope_fractions), keys)

        if mf_factor is not None:
            array = add_mass_fractionation(array, mf_factor, isotope_masses=isotope_masses)
            array = array.normalise(np.sum(isotope_fractions.get(keys), axis=None))

        array = array * val

        for k, v in array.items():
            out[mzkey_map[k.mz()]] += v

    return out


def make_ms_beams(*args, mf_factor=None, fixed_voltage = 10, fixed_key = isopy.keymax, integrations = 100, integration_time=8.389, resistor=1E11,
                  random_seed = None, isotope_fractions=None, isotope_masses=None, **kwargs):
    """
    Simulates a series of measurements with a standard deviation equal to the johnson-nyquist noise
    and counting statistics.

    *args* and *kwargs* are passed to ``isopy.tb.make_ms_array()`` to create the ms array.

    Parameters
    ----------
    args
        Each arg can be either a isotope array, a element key or an isotope key string.
        If arg is an isotope array an exception will be raised if two or more isotopes have the
        same mass number. If arg is an element key string then all isotopes of that element are
        added to the array. If arg is an isotope key string then only that isotope and isotopes
        with the same mass number as any other isotope in the array is added.
    mf_factor
        If given this mass fractionation factor is applied to all isotopes of an element
        prior to being added to the result.
    fixed_voltage
        The voltage of *fixed_key* in the array. The value for all other isotopes in
        the array are adjusted accordingly.
    fixed_key
        If not given then the this defaults to the most abundant isotope. If ``None`` then the sum
        of all isotopes in the array will be set to *fixed_voltage*
    integrations
        The number of simulated measurements. If ``None`` no measurements are simulated and the
        retured array contains the true values.
    integration_time
        The integration time for each simulated measurement.
    resistor
        The resistor used for each measurement. A isotope array or a dictionary can be passed to
        to give different resistor values for different isotopes.
    random_seed
        Must be an integer. Seed given to the random generator.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.
    kwargs
        Each kwarg key must be either an element key string or an isotope key string. The isotope
        fractions added to the array based on the kwarg key are multiplied by the kwarg value.
        Otherwise behaves the same as for *args*.

    Returns
    -------
    IsopyArray
        The simulated measurements.

    Examples
    --------
    >>> isopy.tb.make_ms_beams('pd', ru101=0.01).pprint(nrows=10)
    (row) , 101Ru   , 102Pd   , 104Pd   , 105Pd   , 106Pd    , 108Pd   , 110Pd
    0     , 0.06243 , 0.48869 , 4.14431 , 8.17035 , 9.99984  , 9.68171 , 4.28824
    1     , 0.06241 , 0.48865 , 4.14423 , 8.17033 , 9.99990  , 9.68189 , 4.28823
    2     , 0.06245 , 0.48872 , 4.14423 , 8.17058 , 9.99992  , 9.68178 , 4.28837
    3     , 0.06241 , 0.48868 , 4.14430 , 8.17042 , 10.00005 , 9.68161 , 4.28839
    4     , 0.06240 , 0.48870 , 4.14417 , 8.17053 , 10.00018 , 9.68180 , 4.28834
    ...   , ...     , ...     , ...     , ...     , ...      , ...     , ...
    95    , 0.06244 , 0.48866 , 4.14432 , 8.17056 , 10.00010 , 9.68187 , 4.28827
    96    , 0.06243 , 0.48870 , 4.14410 , 8.17044 , 10.00014 , 9.68184 , 4.28848
    97    , 0.06238 , 0.48867 , 4.14404 , 8.17048 , 9.99992  , 9.68159 , 4.28840
    98    , 0.06245 , 0.48871 , 4.14434 , 8.17031 , 9.99998  , 9.68169 , 4.28829
    99    , 0.06243 , 0.48863 , 4.14438 , 8.17047 , 10.00003 , 9.68145 , 4.28832

    See Also
    --------
    make_ms_array, make_ms_sample
    """

    beams = make_ms_array(*args, **kwargs, mf_factor=mf_factor, isotope_masses=isotope_masses, isotope_fractions=isotope_fractions)
    if beams.size != 1:
        raise ValueError('The constructed ms array has a size larger than one')

    if fixed_voltage is not None:
        beams = beams.normalise(fixed_voltage, fixed_key)

    if integrations is None:
        return beams
    else:
        noise = johnson_nyquist_noise(beams, integration_time=integration_time, resistor=resistor)

        return isopy.random(integrations, list(zip(beams.values(), noise.values())), keys=beams.keys, seed=random_seed)



def make_ms_sample(ms_array, *, fnat = None, fins = None, fixed_voltage = 10, fixed_key = isopy.keymax,
                   blank = None, blank_fixed_voltage = 0.01, blank_fixed_key = isopy.keymax,
                   spike = None, spike_fraction = 0.5,
                   integrations = 100, integration_time = 8.389, resistors = 1E11,
                   random_seed = None,
                   isotope_fractions=None, isotope_masses=None, **interferences):
    """
    Creates a simulated the measurement of a sample with natural and instrumental mass
    fractionation added to the array. The standard deviation of measurements for each isotope
    is equal to the johnson-nyquist noise and counting statistics.

    Parameters
    ----------
    ms_array
        Any object that can be passed to ``make_ms_array`` to returns valid array.
        Also accepts a tuple or a dict which will be unpacked appropriately.
    fnat
        If given, the natural fractionation fractionation factor is applied to the ms_array
        before *interferences* are added to the ms_array.
    fins
        If given, the instrumental mass fractionation factor is applied to the ms_array
        at the same time the *interferences* are added to the ms_array.
    fixed_voltage
        The voltage of *fixed_key* in the array. The value for all other isotopes in
        the array are adjusted accordingly.
    fixed_key
        If not given then the this defaults to the most abundant isotope. If ``None`` then the sum
        of all isotopes in the array will be set to *fixed_voltage*
    blank
        The blank sample to be added to the sample. Can be object that can be singularly passed
        to ``make_ms_array`` which returns valid array. Also accepts a tuple or a dict which will
        be unpacked appropriately.
    blank_fixed_voltage
        The voltage of the *blank_fixed_key* in returned sample that is *blank*.
    blank_fixed_key
        If not given then the this defaults to the most abundant isotope. If ``None`` then the sum
        of all isotopes in the array will be set to *blank_fixed_voltage*
    spike
        If given this spike mixture will be added ms_array after *fnat* but before *fins* and
        *interferences* are added to the array.
    spike_fraction
        The fraction of spike in the final mixture based on the isotopes in *spike*.
    integrations
        The number of simulated measurements. If ``None`` no measurements are simulated and the
        retured array contains the true values.
    integration_time
        The integration time for each simulated measurement.
    resistors
        The resistor used for each measurement. A isotope array or a dictionary can be passed to
        to give different resistor values for different isotopes.
    random_seed
        Must be an integer. Seed given to the random generator.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.
    interferences
        Each kwarg key must be either an element key string or an isotope key string. The isotope
        fractions added to the array based on the kwarg key are multiplied by the kwarg value.
        Otherwise behaves the same as for *args*. A dictionary named interferences can be passed
        for keys that cannot be passed as kwargs, e.g. '137Ba++'.

    Returns
    -------
    IsopyArray
        The simulated measurements.

    Examples
    --------
    >>> isopy.tb.make_ms_sample('pd', ru101=0.01, cd111=0.001, fnat = 0.1, fins=-1.6).pprint(nrows=10)
    (row) , 101Ru   , 102Pd   , 104Pd   , 105Pd   , 106Pd    , 108Pd   , 110Pd   , 111Cd
    0     , 0.06343 , 0.51076 , 4.26008 , 8.28716 , 9.99964  , 9.41352 , 4.06078 , 0.00468
    1     , 0.06340 , 0.51082 , 4.26016 , 8.28714 , 10.00006 , 9.41351 , 4.06064 , 0.00466
    2     , 0.06340 , 0.51083 , 4.26006 , 8.28692 , 9.99981  , 9.41356 , 4.06047 , 0.00470
    3     , 0.06339 , 0.51079 , 4.26018 , 8.28711 , 10.00002 , 9.41349 , 4.06073 , 0.00471
    4     , 0.06341 , 0.51087 , 4.26022 , 8.28713 , 9.99986  , 9.41357 , 4.06062 , 0.00465
    ...   , ...     , ...     , ...     , ...     , ...      , ...     , ...     , ...
    95    , 0.06341 , 0.51080 , 4.26010 , 8.28729 , 9.99992  , 9.41332 , 4.06070 , 0.00465
    96    , 0.06339 , 0.51082 , 4.25994 , 8.28712 , 10.00012 , 9.41364 , 4.06071 , 0.00467
    97    , 0.06339 , 0.51083 , 4.26012 , 8.28728 , 9.99976  , 9.41363 , 4.06069 , 0.00468
    98    , 0.06341 , 0.51078 , 4.26010 , 8.28700 , 9.99994  , 9.41348 , 4.06065 , 0.00469
    99    , 0.06339 , 0.51082 , 4.26010 , 8.28705 , 9.99997  , 9.41386 , 4.06072 , 0.00468


    See Also
    --------
    make_ms_array, make_ms_beams
    """
    ms_array = make_ms_array(ms_array, mf_factor = fnat,
                             isotope_fractions=isotope_fractions, isotope_masses=isotope_masses)

    if spike is not None:
        spike = isopy.asarray(spike)

        spsum = np.sum(spike, axis=1)
        smpsum = np.sum(ms_array.copy(key_eq=spike.keys()), axis = 1)

        spike = spike / spsum * smpsum
        spike = spike * spike_fraction
        ms_array = ms_array * (1 - spike_fraction)

        ms_array = isopy.add(ms_array, spike, 0)

    ms_array = make_ms_array(ms_array, mf_factor = fins,
                             isotope_fractions=isotope_fractions, isotope_masses=isotope_masses,
                             **interferences.pop('interferences', {}), **interferences)

    ms_array = ms_array.normalise(fixed_voltage, fixed_key)

    if blank is not None:
        blank = make_ms_array(blank, mf_factor = fins, isotope_fractions=isotope_fractions,
                              isotope_masses=isotope_masses)

        blank = blank.normalise(blank_fixed_voltage, blank_fixed_key)
        if fixed_key is None:
            sum = isopy.subtract(ms_array, blank, 0, ms_array.keys).sum(axis=None)
            ms_array = ms_array.normalise(sum, None)
        elif isinstance(fixed_key, (str, list, tuple)):
            sum = isopy.subtract(ms_array, blank, 0, fixed_key).sum(axis=None)
            ms_array = ms_array.normalise(sum, fixed_key)
        else:
            sum = isopy.subtract(ms_array, blank, 0, fixed_key(ms_array)).sum(axis=None)
            ms_array = ms_array.normalise(sum, fixed_key(ms_array))

        ms_array = isopy.add(ms_array, blank, 0, ms_array.keys)

    ms_array = make_ms_beams(ms_array, fixed_voltage=None, integrations=integrations,
                             integration_time=integration_time, random_seed=random_seed,
                             resistor=resistors, isotope_fractions=isotope_fractions)

    return ms_array

@core.append_preset_docstring
@core.add_preset(('ppt', 'permil'), extnorm_factor=1000)
@core.add_preset('epsilon', extnorm_factor=1E4)
@core.add_preset(('mu', 'ppm'), extnorm_factor=1E6)
def internal_normalisation(data, mf_ratio, interference_correction=True,
                           extnorm_value = None, extnorm_factor=None,
                           isotope_fractions=None, isotope_masses=None, mf_tol=1E-8):
    """
    A data reduction scheme for internaly normalised data.

    If *interference_correction* is True an interference correction will be applied for all isotopes that are different from the *mf_ratio* numerator
    element. This will be done together with the mass fractionation correction to account for isobaric interferences
    on the *mf_ratio*. If more than one isotope exists for an
    an element the largest isotope is used for the interference correction.

    If *extnorm_value* and/or *extnorm_factor* is given then the returned data is
    externally normalised to these values using the rDelta function. The default value for
    *extnorm_value* is 1 and the default value for *extnorm_factor* is the same as
    *isotope_fractions*, should only one of these values
    be given.

    Parameters
    ----------
    data
        The data to be corrected. Isotope arrays will automatically be converted to ratio arrays.
        Ratio arrays must have a common denominator that is the same as the that of *mf_ratio*.
    mf_ratio
        The data will be internally normalised to this ratio. Must be present in *data*.
    interference_correction
        If True then the data is corrected for isobaric interferences.
    extnorm_value
        If given the result in normalised to this value with *extnorm_factor*. If
        *extnorm_value* is not given a normalisation factor of 1 is used.
    extnorm_factor
        If given and *normalisation_value* is not the result is normalised against
        *isotope_fractions*.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``
    mf_tol
        Only when the difference between the current and the previous mass_fractionation factor
        is below this value is the interference correction assumed to have converged.

    Returns
    -------
    IsopyArray
        Only contains the isotopes of the element with the same element symbol as *mf_ratio*

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd', ru101 = 0.1, cd111 = 0.1, mf_factor=0.1).normalise(10, isopy.keymax)
    >>> array
    (row) , 101Ru   , 102Pd  , 104Pd  , 105Pd  , 106Pd , 108Pd  , 110Pd  , 111Cd
    None  , 0.62089 , 1.5195 , 4.7296 , 8.1258 , 10    , 9.6882 , 4.7396 , 0.46692
    >>> isopy.tb.internal_normalisation(array, '108Pd/105Pd')
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 110Pd/105Pd
    None  , 0.045678    , 0.49888     , 1.2239      , 0.52485
    >>> isopy.tb.internal_normalisation(array, '108Pd/105Pd') / isopy.refval.isotope.fraction
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 110Pd/105Pd
    None  , 1           , 1           , 1           , 1
    >>> isopy.tb.internal_normalisation(array, '108Pd/105Pd', interference_correction=False, extnorm_factor=1)
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 110Pd/105Pd
    None  , 3.12        , 0.16916     , 0.0034253   , 0.10006
    """
    data = isopy.checks.check_array('data', data, 'isotope', 'ratio')
    mf_ratio = isopy.checks.check_type('mf_ratio', mf_ratio, isopy.core.RatioKeyString, coerce=True)
    extnorm_factor = isopy.checks.check_type('normalisation_factor', extnorm_factor, np.float64, str, coerce=True, allow_none=True)
    isotope_fractions = isopy.checks.check_reference_value('isotope_fractions', isotope_fractions, isopy.refval.isotope.fraction)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses, isopy.refval.isotope.mass)

    # Convert the data into a ratio array
    if data.flavour == 'ratio':
        if data.keys.common_denominator is None:
            raise ValueError('data must hav a common denominator')
        elif data.keys.common_denominator != mf_ratio.denominator:
            raise ValueError('data and mf_ratio do not have the same denominator')
        rat = data
    else:
        rat = data.ratio(mf_ratio.denominator)

    if interference_correction is True:
        isobaric_interferences = find_isobaric_interferences(mf_ratio.numerator.element_symbol, data)
    elif  isinstance(interference_correction, dict):
        isobaric_interferences = interference_correction
    else:
        isobaric_interferences = {}

    #Find the initial mass fractionation
    beta = calculate_mass_fractionation_factor(rat, mf_ratio, isotope_fractions=isotope_fractions, isotope_masses=isotope_masses)

    if isobaric_interferences:
        #Do a combined mass fractionation and isobaric interference correction.
        #This can account for isobaric interferences on isotopes in *mf_ratio*
        for i in range(100):
            rat2 = rat
            prev_beta = beta

            rat2 = remove_isobaric_interferences(rat2, isobaric_interferences,
                                                 beta, isotope_fractions=isotope_fractions, isotope_masses=isotope_masses)

            # Calculate the mass fractionation.
            beta = calculate_mass_fractionation_factor(rat2, mf_ratio, isotope_fractions=isotope_fractions, isotope_masses=isotope_masses)

            if np.all(np.abs(beta - prev_beta) < mf_tol):
                break #Beta value has converged so no need for more iterations.
        else:
            raise ValueError('values did not converge after 100 iterations of the interference correction')
    else:
        rat2 = rat

    #Remove the isotopes on interfering elements and the mass bias ratio
    rat = rat2.copy(numerator_element_symbol_eq=mf_ratio.numerator.element_symbol, key_neq=mf_ratio)

    #Correct for mass fractionation
    rat = remove_mass_fractionation(rat, beta, isotope_masses=isotope_masses)

    if extnorm_value is not None:
        if extnorm_factor is None: extnorm_factor = 1
        rat = rDelta(rat, extnorm_value, factor = extnorm_factor)

    elif extnorm_factor is not None:
        rat = rDelta(rat, isotope_fractions, factor = extnorm_factor)

    # Return the corrected data
    return rat


@core.add_preset(('ppt', 'permil'), normalisation_factor=1000)
@core.add_preset('epsilon', normalisation_factor=1E4)
@core.add_preset(('mu', 'ppm'), normalisation_factor=1E6)
@core.renamed_function(internal_normalisation, normalisation_factor='extnorm_factor', normalisation_value='extnorm_value')
def mass_independent_correction(data, mf_ratio,
                           normalisation_value = None, normalisation_factor=None,
                           isotope_fractions=None, isotope_masses=None, mf_tol=1E-8):
    pass

def calculate_mass_fractionation_factor(data, mf_ratio, isotope_fractions=None, isotope_masses=None):
    """
    Calculate the mass fractionation factor for a given ratio in *data*.

    .. math::
        \\alpha = \\ln{( \\frac{r_{n,d}}{R_{n,d}} )} * \\frac{1}{ \\ln{( m_{n,d} } )}

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data
        Fractionated data. :math:`r` in the equation above.
    mf_ratio
        The isotope ratio from which the fractionation factor should be calculated. :math:`n,d` in the equation above.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.

    Returns
    -------
    float
        The fractionation factor for *ratio* in *data*. :math:`\\alpha` in the equation above.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array = isopy.tb.add_mass_fractionation(array, 0.1)
    >>> isopy.tb.calculate_mass_fractionation_factor(array, '108pd/105pd')
    0.09999999999999679

    >>> array = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array = isopy.tb.remove_mass_fractionation(array, 0.15)
    >>> isopy.tb.calculate_mass_fractionation_factor(array, '108pd/105pd')
    -0.1500000000000046

    See Also
    --------
    remove_mass_fractionation
    """
    data = isopy.checks.check_array('data', data, 'isotope', 'ratio')
    mf_ratio = isopy.checks.check_type('ratio', mf_ratio, isopy.core.RatioKeyString, coerce=True)
    isotope_fractions = isopy.checks.check_reference_value('isotope_fractions', isotope_fractions, isopy.refval.isotope.fraction)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses, isopy.refval.isotope.mass)

    if data.flavour == 'isotope':
        data = data.get(mf_ratio.numerator) / data.get(mf_ratio.denominator)
    else:
        data = data.get(mf_ratio)

    return (np.log(data / isotope_fractions.get(mf_ratio)) / np.log(isotope_masses.get(mf_ratio)))


def remove_mass_fractionation(data, fractionation_factor, denom = None, isotope_masses=None):
    """
    Remove exponential mass fractionation from *data*.

    Calculated using:

    .. math::
        R_{n,d} = \\frac{r_{n,d}}{(m_{n,d})^{ \\alpha }}

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data : IsopyArray
        Array containing data to be changed. :math:`r` in the equation above.
    fractionation_factor : float
        Fractionation factor to be applied. :math:`\\alpha` in the equation above
    denom
        The denominator isotope if *data* is an isotope array. If not given then the key with the
        largest median value is used.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.

    Returns
    -------
    IsopyArray
        Will have the same flavour as *data*. :math:`R` in the equation above.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    None  , 0.04568     , 0.49888     , 1.22391     , 1.18495     , 0.52485
    >>> array = isopy.tb.remove_mass_fractionation(array, 0.15)
    >>> array
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    None  , 0.04588     , 0.49960     , 1.22218     , 1.17995     , 0.52120
    >>> isopy.tb.calculate_mass_fractionation_factor(array, '108pd/105pd')
    -0.1500000000000046

    See Also
    --------
    isopy.tb.add_mass_fractionation
    """
    data = isopy.checks.check_array('data', data, 'isotope', 'ratio')
    fractionation_factor = isopy.checks.check_type('fractionation_factor', fractionation_factor, np.float64, np.ndarray,
                                                   coerce=True, coerce_into=[np.float64, np.array], allow_none=True)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses, isopy.refval.isotope.mass)


    if fractionation_factor is None:
        return data
    else:
        if data.flavour == 'isotope':
            if denom is None:
                denom = isopy.keymax(data)
            mass_array = isopy.array(isotope_masses.get(data.keys / denom), data.keys)
        else:
            mass_array = isopy.array(isotope_masses.get(data.keys), data.keys)

        return data / (mass_array ** fractionation_factor)


def add_mass_fractionation(data, fractionation_factor, denom=None, isotope_masses=None):
    """
    Add exponential mass fractionation to *data*.

    Calculated using:

    .. math::
        r_{n,d} = R_{n,d} * (m_{n,d})^{ \\alpha }

    where :math:`n` is the numerator isotope and :math:`d` is the denominator isotope.

    Parameters
    ----------
    data
        Array containing data to which the mass fractionation will be applied. :math:`R` in the equation above.
    fractionation_factor
        Fractionation factor to be applied. :math:`\\alpha` in the equation above
    denom
        The denominator isotope if *data* is an isotope array. If not given then the key with the
        largest median value is used.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.

    Returns
    -------
    IsopyArray
        Will have the same flavour as *data*. :math:`r` in the equation above.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    None  , 0.04568     , 0.49888     , 1.22391     , 1.18495     , 0.52485

    >>> array = isopy.tb.add_mass_fractionation(array, 0.1)
    >>> array
    102Pd/105Pd         , 104Pd/105Pd         , 106Pd/105Pd        , 108Pd/105Pd       , 110Pd/105Pd
    0.04554614408342225 , 0.49840232023288333 , 1.2250738795847462 , 1.188297479483829 , 0.5273039811481615
    >>> isopy.tb.calculate_mass_fractionation_factor(array, '108pd/105pd')
    0.09999999999999679

    See Also
    --------
    isopy.tb.add_mass_fractionation
    """
    data = isopy.checks.check_array('data', data, 'isotope', 'ratio')

    fractionation_factor = isopy.checks.check_type('fractionation_factor', fractionation_factor, np.float64, np.ndarray,
                                                   coerce=True, coerce_into=[np.float64, np.array],allow_none=True)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses, isopy.refval.isotope.mass)

    if fractionation_factor is None:
        return data
    else:
        if data.flavour == 'isotope':
            if denom is None:
                denom = isopy.keymax(data)
            mass_array = isopy.array(isotope_masses.get(data.keys / denom), keys=data.keys)
        else:
            mass_array = isopy.array(isotope_masses.get(data.keys), data.keys)

        return  data * (mass_array ** fractionation_factor)


def find_isobaric_interferences(main, interferences=None) -> isopy.IsopyDict:
    """
    Find isobaric interferences for all isotopes in *main* for each element in *interferences*.

    Parameters
    ----------
    main
        A string, sequence of string or any object compatible with ``askeystring`` that returns a
        element or isotope key list. For element keys it will find interferences for all isotopes
        in *interferences* with that element symbol. If there are no isotopes in *interferences*
        with that element symbol it will find the isobaric interferences for all naturally occurring
        isotopes of that element.
    interferences
        A string, sequence of string or any object compatible with ``askeystring`` that returns a
        element or isotope key list. It will find the isobaric interferences for all naturally
        occurring isotopes of elements in *interferences*. If not given then it will find the
        isobaric interferences for isotopes in main of all naturally occurring elements.

    Returns
    -------
    isobaric_interferences
        A dictionary containing the all isobaric interferences of isotopes in *main* for elements
        in *interferences*.

    Examples
    --------
    >>> isopy.tb.find_isobaric_interferences('pd', ('ru', 'cd'))
    IsopyDict(default_value = (), readonly = False,
    {"Ru": IsotopeKeyList('102Pd', '104Pd')
    "Cd": IsotopeKeyList('106Pd', '108Pd', '110Pd')})

    >>> array = isopy.tb.make_ms_array('pd', '101ru', '111cd')
    >>> isopy.tb.find_isobaric_interferences('pd', array)
    IsopyDict(default_value = (), readonly = False,
    {"Ru": IsotopeKeyList('102Pd', '104Pd')
    "Cd": IsotopeKeyList('106Pd', '108Pd', '110Pd')})

    >>> isopy.tb.find_isobaric_interferences('ce')
    IsopyDict(default_value = (), readonly = False,
    {"Xe": IsotopeKeyList('136Ce')
    "Ba": IsotopeKeyList('136Ce', '138Ce')
    "La": IsotopeKeyList('138Ce')
    "Nd": IsotopeKeyList('142Ce')})

    >>> isopy.tb.find_isobaric_interferences('ce138')
    IsopyDict(default_value = (), readonly = False,
    {"Ba": IsotopeKeyList('138Ce')
    "La": IsotopeKeyList('138Ce')})
    """
    main = isopy.askeylist(main).flatten(ignore_duplicates=True)
    if not isinstance(main, (isopy.ElementKeyList, isopy.IsotopeKeyList)):
        raise ValueError(f'main must be element or isotope key list not {type(main).__name__}')

    if interferences is not None:
        interferences = isopy.askeylist(interferences).flatten(ignore_duplicates=True)
        if not isinstance(interferences, (isopy.ElementKeyList, isopy.IsotopeKeyList)):
            raise ValueError(
                f'interference must be element or isotope key list not {type(interferences).__name__}')

        if isinstance(interferences, isopy.ElementKeyList):
            interference_isotopes = isopy.IsotopeKeyList()
            for element in interferences:
                isotopes = isopy.refval.element.isotopes.get(element, None)
                if isotopes is None: continue
                isotopes = isotopes.set_charges(element.charge)
                interference_isotopes += isotopes
        else:
            interference_isotopes = interferences
    else:
        interference_isotopes = isopy.IsotopeKeyList()

    if isinstance(main, isopy.ElementKeyList):
        main_isotopes = isopy.IsotopeKeyList()
        for element in main:
            if element in interference_isotopes.element_symbols:
                main_isotopes += interference_isotopes.filter(element_symbol_eq=element)
            else:
                main_isotopes += isopy.refval.element.isotopes.get(element, tuple())
    else:
        main_isotopes = main

    if interferences is None:
        interference_isotopes = isopy.IsotopeKeyList()
        for mass in main_isotopes.mass_numbers:
            interference_isotopes += isopy.refval.mass.isotopes[mass]

    interference_elements = interference_isotopes.element_symbols.flatten(ignore_duplicates=True)

    result = isopy.IsopyDict(default_value=tuple())
    for element in interference_elements:
        interferences = isopy.refval.element.isotopes.get(element, None)
        if interferences is None: continue
        interferences = interferences.set_charges(element.charge)
        main = main_isotopes.filter(element_symbol_neq=element,
                                    mz_eq=interferences.mz())
        if len(main) > 0:
            result[element] = main
    return result


def remove_isobaric_interferences(data, isobaric_interferences, mf_factor=None,
                                   isotope_fractions=None, isotope_masses=None):
    """
    Remove all isobaric interferences for a given element(s).

    Parameters
    ----------
    data
        The data that isobaric interferences should be removed from.
    isobaric_interferences
        A dictionary mapping either the element or the isotope to be the isotopes that
        this element has isobaric interferences on that you wish to correct for. If the key is
        an element key string then the isotope of that element in the array with the largest
        signal will be used for the correction.
    mf_factor
        If given, this mass fractionation factor is applied to the values for for the correction.
    isotope_fractions
        Reference value for the isotope fractions of different elements.
        Defaults to ``isopy.refval.isotope.abundance``.
    isotope_masses
        Reference value for the isotope masses of different elements.
        Defaults to ``isopy.refval.isotope.mass``.

    Returns
    -------
    IsopyArray
        The interference corrected array

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd', ru101 = 0.1, cd111 = 0.1)
    >>> array
    (row) , 101Ru   , 102Pd    , 104Pd   , 105Pd  , 106Pd   , 108Pd   , 110Pd   , 111Cd
    None  , 0.01706 , 0.041752 , 0.13002 , 0.2233 , 0.27455 , 0.26549 , 0.12968 , 0.012804
    >>> interferences = isopy.tb.find_isobaric_interferences('pd', array)
    >>> interferences
    IsopyDict(default_value = (), readonly = False,
    {"Ru": IsotopeKeyList('102Pd', '104Pd')
    "Cd": IsotopeKeyList('106Pd', '108Pd', '110Pd')})
    >>> isopy.tb.remove_isobaric_interferences(array, interferences)
    (row) , 101Ru , 102Pd  , 104Pd  , 105Pd  , 106Pd  , 108Pd  , 110Pd  , 111Cd
    None  , 0     , 0.0102 , 0.1114 , 0.2233 , 0.2733 , 0.2646 , 0.1172 , 0
    >>> isopy.tb.make_ms_array('pd', ru101 = 0, cd111 = 0)
    (row) , 101Ru , 102Pd  , 104Pd  , 105Pd  , 106Pd  , 108Pd  , 110Pd  , 111Cd
    None  , 0     , 0.0102 , 0.1114 , 0.2233 , 0.2733 , 0.2646 , 0.1172 , 0

    >>> array = isopy.tb.make_ms_array('pd', ru101 = 0.1, cd111 = 0.1).ratio('106pd')
    >>> array
    (row) , 101Ru/106Pd , 102Pd/106Pd , 104Pd/106Pd , 105Pd/106Pd , 108Pd/106Pd , 110Pd/106Pd , 111Cd/106Pd
    None  , 0.062138    , 0.15207     , 0.47358     , 0.81333     , 0.967       , 0.47236     , 0.046636
    >>> interferences = isopy.tb.find_isobaric_interferences('pd', array)
    >>> interferences
    IsopyDict(default_value = (), readonly = False,
    {"Ru": IsotopeKeyList('102Pd', '104Pd')
    "Cd": IsotopeKeyList('106Pd', '108Pd', '110Pd')})
    >>> isopy.tb.remove_isobaric_interferences(array, interferences)
    (row) , 101Ru/106Pd , 102Pd/106Pd , 104Pd/106Pd , 105Pd/106Pd , 108Pd/106Pd , 110Pd/106Pd , 111Cd/106Pd
    None  , 0           , 0.037322    , 0.40761     , 0.81705     , 0.96817     , 0.42883     , 0
    >>> isopy.tb.make_ms_array('pd', ru101 = 0, cd111 = 0).ratio('106pd')
    (row) , 101Ru/106Pd , 102Pd/106Pd , 104Pd/106Pd , 105Pd/106Pd , 108Pd/106Pd , 110Pd/106Pd , 111Cd/106Pd
    None  , 0           , 0.037322    , 0.40761     , 0.81705     , 0.96817     , 0.42883     , 0
    """

    data = isopy.checks.check_array('data', data, 'isotope', 'ratio')

    mf_factor = isopy.checks.check_type('mf_factor', mf_factor, np.float64, np.ndarray, coerce=True,
                                        coerce_into=[np.float64, np.array], allow_none=True)
    isotope_fractions = isopy.checks.check_reference_value('isotope_fractions',
                                                            isotope_fractions,
                                                            isopy.refval.isotope.fraction)
    isotope_masses = isopy.checks.check_reference_value('isotope_masses', isotope_masses,
                                                        isopy.refval.isotope.mass)

    # Get the information we need to make the correction that works
    # for both isotope arrays and ratio arrays
    if data.flavour == 'ratio':
        keys = data.keys()
        numer = keys.numerators
        denom = keys.common_denominator
    elif data.flavour == 'isotope':
        keys = data.keys()
        numer = keys
        denom = None

    if isinstance(isobaric_interferences, dict) and not isinstance(isobaric_interferences, core.IsopyDict):
        isobaric_interferences = core.IsopyDict(isobaric_interferences)
    elif not isinstance(isobaric_interferences, isopy.IsopyDict):
        raise ValueError('isobaric_interferences must be a dict')

    out = data.copy()

    for interference_element, correct_isotopes in isobaric_interferences.items():
        #If the key of the dictionary is an isotope use this isotope to make the correction
        if isinstance(interference_element, isopy.IsotopeKeyString):
            if interference_element not in numer:
                raise KeyError(f'isotope {interference_element} not in data')
            interference_isotope = interference_element
            interference_element = interference_isotope.element_symbol

        #If the key is an element key string use the isotope of that element in the array that has the largest signal.
        else:
            if data.flavour == 'ratio':
                interference_isotope = isopy.keymax(data.copy(numerator_element_symbol_eq=interference_element))
                interference_isotope = interference_isotope.numerator
            else:
                interference_isotope = isopy.keymax(data.copy(element_symbol_eq=interference_element))

        # Get the abundances of all isotopes of the interfering element
        inf_data = isotope_fractions.copy(element_symbol_eq=interference_isotope.element_symbol.basekey)
        inf_data  = {key.set_charge(interference_isotope.charge): value for key, value in inf_data.items()}
        inf_data = isopy.array(inf_data)

        # Turn into a ratio relative to interference isotope
        inf_data = inf_data.ratio(interference_isotope, remove_denominator=False)

        # Account for mass fractionation of *mf_factor* is given
        if mf_factor is not None:
            inf_data = add_mass_fractionation(inf_data, mf_factor, isotope_masses=isotope_masses)

        # Scale relative to the measured value of *isotope*
        inf_data = inf_data * data[keys[numer.index(interference_isotope)]]

        # Convert to a mass array for easy lookup later
        inf_data = {key.numerator.mz(): value for key, value in inf_data.items()}

        # Loop through each key and remove interference
        for i, key in enumerate(keys):
            if (numer[i] in correct_isotopes or numer[i].element_symbol == interference_element):
                out[key] = out[key] - inf_data.get(numer[i].mz(), 0)

        # If *data* is a ratio array then correct for any interference on the denominator
        if denom is not None and denom in correct_isotopes:
            out = out / (1 - inf_data.get(denom.mz(), 0))

    return out

@core.append_preset_docstring
@core.add_preset(('ppt', 'permil'), factor=1000)
@core.add_preset('epsilon', factor=1E4)
@core.add_preset(('mu', 'ppm'), factor=1E6)
def rDelta(data, reference_data, factor=1, deviations=1):
    """
    Externally normalise data to the given reference values.

    .. math::
        \Delta^{r} \\textrm{normalised} = (\\frac{\\textrm{data}} {\\textrm{reference data}} - \\textrm{deviations} ) * \\textrm{factor}


    Parameters
    ----------
    data
        Data to be normalised
    reference_data
        The reference values used to normalise the data. If a reference values contains more than
        one value the mean is used. Multiple values can be passed as a tuple in which case the
        mean of those values is used.
    factor
        The multiplication factor to be applied to *data* during the normalisation.
    deviations
        ``1`` if *reference_data* should be subtracted from *data*.

    Examples
    --------
    >>> ref = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array = isopy.tb.make_ms_sample('pd', fins=1, fixed_voltage=0.1).ratio('105pd')
    >>> norm = isopy.tb.rDelta(array, ref)
    >>> norm
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    0     , -28.60967   , -9.54006    , 9.50745     , 28.57674    , 47.70528
    1     , -28.76605   , -9.52060    , 9.53867     , 28.61274    , 47.69063
    2     , -28.73309   , -9.56115    , 9.51017     , 28.57170    , 47.63513
    3     , -28.53972   , -9.60582    , 9.48726     , 28.57055    , 47.60977
    4     , -28.69282   , -9.59705    , 9.50470     , 28.55289    , 47.65280
    ...   , ...         , ...         , ...         , ...         , ...
    95    , -28.50008   , -9.60504    , 9.52317     , 28.56964    , 47.66661
    96    , -28.64389   , -9.47620    , 9.52398     , 28.59006    , 47.70599
    97    , -28.64383   , -9.55584    , 9.48523     , 28.58221    , 47.67802
    98    , -28.64546   , -9.53501    , 9.52313     , 28.59109    , 47.66902
    99    , -28.67808   , -9.52044    , 9.51687     , 28.57950    , 47.62100

    >>> isopy.tb.rDelta(isopy.sd2(array), ref, factor=1000, deviations=True)
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    None  , 0.17452     , 0.05670     , 0.03890     , 0.03899     , 0.04946
    >>> isopy.sd2(norm)
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    None  , 0.17452     , 0.05670     , 0.03890     , 0.03899     , 0.04946
    """
    data = isopy.checks.check_type('data', data, isopy.core.IsopyArray, coerce=True, coerce_into=isopy.core.asarray)
    factor = isopy.checks.check_type('factor', factor, np.float64, coerce=True)
    deviations = isopy.checks.check_type('deviations', deviations, np.float64, coerce=True)

    reference_data = _combine_reference_values(reference_data)

    new = data / reference_data
    new = new - deviations
    new = new * factor

    return new

@core.append_preset_docstring
@core.add_preset(('ppt', 'permil'), factor=1000)
@core.add_preset('epsilon', factor=1E4)
@core.add_preset(('mu', 'ppm'), factor=1E6)
def inverse_rDelta(data, reference_data, factor=1, deviations=1):
    """
    Denormalise data to the given reference values.

    .. math::
        \\textrm{denormalised data} = (\\frac{\Delta^{r} \\textrm{data}} {\\textrm{factor}} + \\textrm{deviations}) * \\textrm{reference data}


    Parameters
    ----------
    data
        Normalised data to be denormalised
    reference_data
        The reference values used to denormalise the data. If multiple values are passed
        the mean of all the reference values are used. If a reference values contains more than
        one value the mean is used.
    factor
        The multiplication factor applied to *data* during the normalisation.
    deviations
        ``1`` if *reference_data* should be added to the denormalised data.


    Examples
    --------
    >>> ref = isopy.tb.make_ms_array('pd').ratio('105pd')
    >>> array = isopy.tb.make_ms_sample('pd', fins=1).ratio('105pd')
    >>> norm = isopy.toolbox.isotope.rDelta(array, ref, 1000)
    >>> denorm = isopy.tb.inverse_rDelta(norm, ref, 1000)
    >>> denorm
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    0     , 0.04437     , 0.49412     , 1.23554     , 1.21881     , 0.54985
    1     , 0.04437     , 0.49413     , 1.23556     , 1.21882     , 0.54986
    2     , 0.04437     , 0.49412     , 1.23558     , 1.21887     , 0.54988
    3     , 0.04438     , 0.49413     , 1.23560     , 1.21887     , 0.54988
    4     , 0.04437     , 0.49413     , 1.23553     , 1.21879     , 0.54985
    ...   , ...         , ...         , ...         , ...         , ...
    95    , 0.04436     , 0.49415     , 1.23562     , 1.21883     , 0.54989
    96    , 0.04438     , 0.49412     , 1.23558     , 1.21884     , 0.54988
    97    , 0.04437     , 0.49412     , 1.23555     , 1.21888     , 0.54986
    98    , 0.04437     , 0.49411     , 1.23557     , 1.21880     , 0.54986
    99    , 0.04438     , 0.49410     , 1.23558     , 1.21884     , 0.54987
    >>> denorm/array
    (row) , 102Pd/105Pd , 104Pd/105Pd , 106Pd/105Pd , 108Pd/105Pd , 110Pd/105Pd
    0     , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    1     , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    2     , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    3     , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    4     , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    ...   , ...         , ...         , ...         , ...         , ...
    95    , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    96    , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    97    , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    98    , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    99    , 1.00000     , 1.00000     , 1.00000     , 1.00000     , 1.00000
    """
    data = isopy.checks.check_type('data', data, isopy.core.IsopyArray, coerce=True, coerce_into=isopy.core.asarray)
    factor = isopy.checks.check_type('factor', factor, np.float64, coerce=True)
    deviations = isopy.checks.check_type('deviations', deviations, np.float64, coerce=True)

    reference_data = _combine_reference_values(reference_data)

    new = data / factor
    new = new + deviations
    new = new * reference_data

    return new


def _combine_reference_values(values):
    if isinstance(values, dict):
        return values

    if not isinstance(values, (list, tuple)):
        values = (values,)
    else:
        values = values

    out = []
    for v in values:
        v = isopy.asarray(v)
        if v.size > 1:
            v = np.nanmean(v)
        out.append(v)

    if len(out) == 1:
        return isopy.ScalarDict(out[0])
    else:
        return isopy.ScalarDict(np.mean(isopy.concatenate(out)))

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=isopy.sd)
@core.add_preset('sd2', cval=np.mean, pmval=isopy.sd2)
@core.add_preset('sd3', cval=np.mean, pmval=isopy.sd3)
@core.add_preset('se', cval=np.mean, pmval=isopy.se)
@core.add_preset('se2', cval=np.mean, pmval=isopy.se2)
@core.add_preset('se3', cval=np.mean, pmval=isopy.se3)
@core.add_preset('mad', cval=np.median, pmval=isopy.mad)
@core.add_preset('mad2', cval=np.median, pmval=isopy.mad2)
@core.add_preset('mad3', cval=np.median, pmval=isopy.mad3)
def find_outliers(data, cval = np.median, pmval=isopy.mad3, axis = None, invert=False):
    """
    Find all outliers in data.

    If *invert* is ``False`` the result will be an array where outliers
    are marked with ``True`` and everything else ``False``. If *invert*
    is ``True`` the result is inverted so outliers are marked with
    ``False``.

    Parameters
    ----------
    data : isopy_array_like
        The outliers will be calculated for each column in *data*.
    cval : scalar, Callable
        Either the center value or a function that returns the center
        value when called with *data*.
    pmval : scalar, Callable
        Either the uncertainty value or a function that returns the
        uncertainty when called with *data*.
    axis : {0, 1}, Optional
        If not given then an array with each individual outlier marked is returned. Otherwise
        ``np.any(outliers, axis)`` is returned.
    invert : bool
        If ``True`` the output is inverted.

    Examples
    --------
    >>> array = isopy.tb.make_ms_sample('pd', integrations=10)
    >>> array['pd102'][:2] *= 2
    >>> array['pd110'][-1] *= 2
    >>> isopy.tb.find_outliers(array)
    (row) , 102Pd , 104Pd , 105Pd , 106Pd , 108Pd , 110Pd
    0     , True  , False , True  , False , False , False
    1     , True  , False , False , False , False , False
    2     , False , False , False , False , False , False
    3     , False , False , False , False , False , False
    4     , False , False , False , False , False , False
    5     , False , False , False , False , False , False
    6     , False , False , False , False , False , False
    7     , False , False , False , False , False , False
    8     , False , False , False , False , False , False
    9     , False , False , False , False , False , True

    >>> isopy.tb.find_outliers(array, axis=1)
    array([ True,  True, False, False, False, False, False, False, False, True])

    >>> isopy.tb.find_outliers(array, np.mean, isopy.sd2)
    (row) , 102Pd , 104Pd , 105Pd , 106Pd , 108Pd , 110Pd
    0     , False , False , True  , False , False , False
    1     , False , False , False , False , False , False
    2     , False , False , False , False , False , False
    3     , False , False , False , False , False , False
    4     , False , False , False , False , False , False
    5     , False , False , False , False , False , False
    6     , False , False , False , False , False , False
    7     , False , False , False , False , False , False
    8     , False , False , False , False , False , False
    9     , False , False , False , False , False , True

    >>> isopy.tb.find_outliers(array, np.mean(array), isopy.sd2(array))
    (row) , 102Pd , 104Pd , 105Pd , 106Pd , 108Pd , 110Pd
    0     , False , False , True  , False , False , False
    1     , False , False , False , False , False , False
    2     , False , False , False , False , False , False
    3     , False , False , False , False , False , False
    4     , False , False , False , False , False , False
    5     , False , False , False , False , False , False
    6     , False , False , False , False , False , False
    7     , False , False , False , False , False , False
    8     , False , False , False , False , False , False
    9     , False , False , False , False , False , True
    """
    axis = isopy.checks.check_type('axis', axis, int, allow_none=True)

    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    outliers = (data > (cval + pmval)) + (data < (cval - pmval))

    if axis is not None:
        outliers = np.any(outliers, axis=axis)

    if invert:
        outliers = np.invert(outliers)

    return outliers

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=isopy.sd)
@core.add_preset('sd2', cval=np.mean, pmval=isopy.sd2)
@core.add_preset('sd3', cval=np.mean, pmval=isopy.sd3)
@core.add_preset('se', cval=np.mean, pmval=isopy.se)
@core.add_preset('se2', cval=np.mean, pmval=isopy.se2)
@core.add_preset('se3', cval=np.mean, pmval=isopy.se3)
@core.add_preset('mad', cval=np.median, pmval=isopy.mad)
@core.add_preset('mad2', cval=np.median, pmval=isopy.mad2)
@core.add_preset('mad3', cval=np.median, pmval=isopy.mad3)
def upper_limit(data, cval=np.median, pmval=isopy.mad3):
    """
    Calculate the upper limit of the uncertainty on *data* as ``cval + pmval``.


    Parameters
    ----------
    data
        The data on which the limit will be calculated
    cval
        The centre value. Can either be a scalar value or a function that returns the centre value.
    pmval
        The uncertainty around *cval*. Can either be a scalar value or a function that returns the uncertainty.
    """
    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    return cval + pmval

@core.append_preset_docstring
@core.add_preset('sd', cval=np.mean, pmval=isopy.sd)
@core.add_preset('sd2', cval=np.mean, pmval=isopy.sd2)
@core.add_preset('sd3', cval=np.mean, pmval=isopy.sd3)
@core.add_preset('se', cval=np.mean, pmval=isopy.se)
@core.add_preset('se2', cval=np.mean, pmval=isopy.se2)
@core.add_preset('se3', cval=np.mean, pmval=isopy.se3)
@core.add_preset('mad', cval=np.median, pmval=isopy.mad)
@core.add_preset('mad2', cval=np.median, pmval=isopy.mad2)
@core.add_preset('mad3', cval=np.median, pmval=isopy.mad3)
def lower_limit(data, cval=np.median, pmval=isopy.mad3):
    """
    Calculate the lower limit of the uncertainty on *data* as ``cval + pmval``.

    Parameters
    ----------
    data
        The data on which the limit will be calculated
    cval
        The centre value. Can either be a scalar value or a function that returns the centre value.
    pmval
        The uncertainty around *cval*. Can either be a scalar value or a function that returns the uncertainty.
    """
    if callable(cval):
        cval = cval(data)
    if callable(pmval):
        pmval = pmval(data)
    pmval = np.abs(pmval)

    return cval - pmval


import isopy
from isopy import core
from isopy import io
import numpy as np
import functools
import os

__all__ = ['refval']

########################
### Reference Values ###
########################

def _load_RV_values(filename, datatype=None):
    filepath = os.path.join(os.path.dirname(__file__), 'referencedata', f'{filename}.csv')
    data = io.read_csv(filepath)

    if datatype is not None:
        data =  {key: datatype(value[0]) for key, value in data.items()}
    else:
        data = {key: value[0] for key, value in data.items()}
    return data

# Assumes all default and reference values are properties
def is_default_value(func):
    setattr(func.fget, '_defval', True)
    setattr(func.fget, '_descr', func.fget.__doc__.lstrip('\n').split('\n', 1)[0].strip())
    return func
    
def is_reference_value(func):
    setattr(func.fget, '_refval', True)
    setattr(func.fget, '_descr', func.fget.__doc__.lstrip('\n').split('\n', 1)[0].strip())
    return func

class RefValGroup:
    def __repr__(self):
        defval = []
        refval = []
        for name, item in self.__class__.__dict__.items():
            if type(item) is property:
                if getattr(item.fget, '_defval', False):
                    defval.append(f'{name} - {item.fget._descr}')
                if getattr(item.fget, '_refval', False):
                    refval.append(f'{name} - {item.fget._descr}')
        
        string = ''
        if defval:
            string += 'This group contains the following default values:\n\t'
            string += '\n\t'.join(defval)
        if string: string += '\n\n'
        string += 'This group contains the following reference values:\n\t'
        string += '\n\t'.join(refval)

        return string

    def ls(self):
        """
        Returns a list will all reference values in this group.
        """
        defval = []
        refval = []
        for name, item in self.__class__.__dict__.items():
            if type(item) is property:
                if getattr(item.fget, '_defval', False):
                    defval.append(name)
                if getattr(item.fget, '_refval', False):
                    refval.append(name)
            
        return defval + refval


class mass(RefValGroup):
    @is_reference_value        
    @core.cached_property
    def isotopes(self):
        """
        Dictionary containing all naturally isotopes with a given mass number.

        The dictionary is constructed from the isotopes in
        :attr:`isotope.best_abundance_measurement_M16`.

        The ``get()`` method of this dictionary will
        return an empty :class:`IsotopeKeyList` for absent keys.

        Examples
        --------
        >>> isopy.refval.mass.isotopes[40]
        IsotopeKeyList('40Ar', '40K', '40Ca')

        >>> isopy.refval.mass.isotopes.get(96)
        IsotopeKeyList('96Zr', '96Mo', '96Ru')
        """

        fraction = isopy.refval.isotope.best_measurement_fraction_M16
        isotopes = {f'{mass}':
                    isopy.IsotopeKeyList([k for k in fraction if k._filter(mass_number = {'key_eq': (mass,)})])
                    for mass in range(1, 239)}
        return core.IsopyDict(**{key: value for key, value in isotopes.items() if len(value) > 0},
                               default_value=isopy.IsotopeKeyList(), readonly=True)


#TODO use best isotope and isotope weight to calculate atomic weight
class element(RefValGroup):
    @is_reference_value
    @core.cached_property
    def isotopes(self):
        """
        Dictionary containing all naturally occurring isotopes for each element symbol.

        The dictionary is constructed from the isotopes in
        :attr:`isotope.best_abundance_measurement_M16`.

        The ``get()`` method of this dictionary will
        return an empty :class:`IsotopeKeyList` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.isotopes['pd']
        IsotopeKeyList('102Pd', '104Pd', '105Pd', '106Pd', '108Pd', '110Pd')

        >>> isopy.refval.element.isotopes.get('ge')
        IsotopeKeyList('70Ge', '72Ge', '73Ge', '74Ge', '76Ge')
        """

        fraction = isopy.refval.isotope.best_measurement_fraction_M16
        return core.IsopyDict(**{symbol: isopy.IsotopeKeyList([k for k in fraction if k._filter(element_symbol = {'key_eq': (symbol,)})])
                               for symbol in self.all_symbols}, default_value=isopy.IsotopeKeyList(), readonly=True)
    
    @is_reference_value
    @core.cached_property
    def all_symbols(self):
        """
        A tuple of all the element symbols.

        Examples
        --------
        >>> isopy.refval.element.all_symbols[:5] # first 5 element symbols
        (ElementKeyString('H'), ElementKeyString('He'), ElementKeyString('Li'),
         ElementKeyString('Be'), ElementKeyString('B'))
        """
        return tuple(self.symbol_name.keys())
    
    @is_reference_value
    @is_default_value  
    @core.cached_property
    def symbol_name(self):
        """
        Dictionary containing the full element name mapped to the element symbol.

        The first letter of the element name is capitalised.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.symbol_name['pd']
        'Palladium'

        >>> isopy.refval.element.symbol_name.get('ge')
        'Germanium'
        """
        data = {isopy.ElementKeyString(key, allow_reformatting=False): value for key, value in _load_RV_values('element_symbol_name').items()}
        return core.IsopyDict(data, readonly = True)
        
    @is_reference_value
    @core.cached_property
    def name_symbol(self):
        """
        Dictionary containing the element symbol mapped to the full element name.

        The first letter of the element name must be capitalised.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.symbol_name['Palladium']
        ElementKeyString('Pd')

        >>> isopy.refval.element.symbol_name.get('Germanium')
        ElementKeyString('Ge')
        """
        return dict({name: symbol for symbol, name in self.symbol_name.items()}, readonly = True)

    @is_reference_value
    @core.cached_property
    def atomic_weight(self):
        """
        Dictionary containing the atomic weight for each element symbol.

        The atomic weights are calculated using the isotopic abundances from
        :attr:`isotope.best_abundance_measurement_M16` and the isotopic masses from
        :attr:`isotope.isotope.mass_W17`.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.atomic_weight['pd']
        106.41532788648

        >>> isopy.refval.element.atomic_weight.get('ge')
        72.6295890304831
        """
        weights = {}
        # Isotopes is also taken from best_measurement_fraction_M16
        for element, isotopes in self.isotopes.items():
            weights[element] = np.sum([refval.isotope.mass_W17.get(isotope) *
                                       refval.isotope.best_measurement_fraction_M16.get(isotope)
                                       for isotope in isotopes])
        return core.IsopyDict(**weights, readonly=True)
        
    @is_reference_value
    @core.cached_property
    def atomic_number(self):
        """
        Dictionary containing the atomic number for each element symbol.

        The ``get()`` method of this dictionary will
        return ``None`` for absent keys.

        Examples
        --------
        >>> isopy.refval.element.atomic_number['pd']
        46

        >>> isopy.refval.element.atomic_number.get('ge')
        32
        """
        return core.IsopyDict(_load_RV_values('element_atomic_number', int), readonly = True)
        
    @is_reference_value
    @core.cached_property
    def initial_solar_system_abundance_L09(self):
        """
        Dictionary containing the element abundance of the initial solar system composition (normalized to N(Si) = 10^6 atoms) from Lodders et al. (2019).

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.
        """
        return core.ScalarDict(_load_RV_values('element_initial_solar_system_abundance_L09', np.float64), default_value=np.nan, readonly = True)

class isotope(RefValGroup):
    def __init__(self):
        self.__mass = None
        self.__fraction = None
    
    @is_default_value
    @property
    def mass(self):
        """
        Dictionary containing the default mass of each isotope.

        This is the default dictionary used in functions where the isotope mass is required.

        This dictionary is initially a copy of :attr:`isotope.mass_W17` but can be changed by the
        user. **Note** the default values used can change in future versions of isopy as newer
        values become avaliable.

        Examples
        --------
        >>> isopy.refval.isotope.mass['pd105']
        104.9050795

        >>> isopy.refval.isotope.mass.get('ge76')
        75.92140273

        >>> isopy.refval.isotope.mass.get('pd108/pd105')
        1.0285859589859039
        """
        if not self.__mass: self.__mass = self.mass_W17
        return self.__mass

    @mass.setter
    def mass(self, value):
        if not isinstance(value, core.IsopyDict):
            raise TypeError('attribute must be a dictionary')
        self.__mass = value
    
    @is_default_value
    @property
    def fraction(self):
        """
        Dictionary containing the default fraction of each isotope.

        This is the default dictionary used in functions where the isotope abundance is required.

        This dictionary is initially a copy of :attr:`isotope.best_measurement_fracton_M16`
        but can be changed by the user. **Note** the default values used can change in future
        versions of isopy as newer values become avaliable.

        Examples
        --------
        >>> isopy.refval.isotope.best_measurement_fraction_M16['pd105']
        0.2233

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('ge76')
        0.07745

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('pd108/pd105')
        1.1849529780564263
        """
        if not self.__fraction: self.__fraction = self.best_measurement_fraction_M16
        return self.__fraction

    @fraction.setter
    def fraction(self, value):
        if not isinstance(value, core.IsopyDict):
            raise TypeError('attribute must be a dictionary')
        self.__fraction = value
    
    @is_reference_value
    @core.cached_property
    def mass_W17(self):
        """
        Dictionary containing isotope mass of each isotope from Wang et al. (2016).

        Reference: `Wang et al. (2016) <https://doi.org/10.1088/1674-1137/41/3/030003>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.mass_W17['pd105']
        104.9050795

        >>> isopy.refval.isotope.mass_W17.get('ge76')
        75.92140273

        >>> isopy.refval.isotope.mass_W17.get('pd108/pd105')
        1.0285859589859039
        """
        return core.ScalarDict(_load_RV_values('isotope_mass_W17', np.float64), default_value=np.nan, readonly = True)

    @is_reference_value
    @core.cached_property
    def mass_number(self):
        """
        Dictionary containing mass number of each isotope.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.mass_number['pd105']
        105

        >>> isopy.refval.isotope.mass_number.get('ge76')
        76

        >>> isopy.refval.isotope.mass_number.get('pd108/pd105')
        1.0285714285714285
        """
        return core.ScalarDict({key: int(key.mass_number) for key in self.mass_W17},
                               default_value=np.nan, readonly=True)
    
    @is_reference_value
    @core.cached_property
    def best_measurement_fraction_M16(self):
        """
        Dictionary containing the isotope fraction from the best avaliable measurement from Meija et al. (2016).

        Reference: `Meija et al. (2016) <https://doi.org/10.1515/pac-2015-0503>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.best_measurement_fraction_M16['pd105']
        0.2233

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('ge76')
        0.07745

        >>> isopy.refval.isotope.best_measurement_fraction_M16.get('pd108/pd105')
        1.1849529780564263
        """
        return core.ScalarDict(_load_RV_values('isotope_best_measurement_fraction_M16', np.float64), default_value=np.nan, readonly = True)
        
    @is_reference_value
    @core.cached_property
    def initial_solar_system_fraction_L09(self):
        """
        Dictionary containing the isotope fraction of the inital solar system composition from Lodders et al. (2019).

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_fraction_L09['pd105']
        0.2233

        >>> isopy.refval.isotope.initial_solar_system_fraction_L09.get('ge76')
        0.07444

        >>> isopy.refval.isotope.initial_solar_system_fraction_L09.get('pd108/pd105')
        1.1849529780564263
        """
        return core.ScalarDict(_load_RV_values('isotope_initial_solar_system_fraction_L09', np.float64), default_value = np.nan, readonly = True)
        
    @is_reference_value
    @core.cached_property
    def initial_solar_system_abundance_L09(self):
        """
        Dictionary containing the isotope abundance of the inital solar system composition (normalized to N(Si) = 10^6 atoms) from Lodders et al. (2019).

        **Note** These values are not directly taken from the table but calculated from the elemental abundances and
        the isotope fractions given the table. This ensures consistency between all three reference values. Discrepancies
        between the calculated isotope abundances and those listen in the table are due to rounding errors.

        Reference: `Lodders et al. (2019) <http://materials.springer.com/lb/docs/sm_lbs_978-3-540-88055-4_34>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_abundance_L09['pd105']
        0.303688

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('ge76')
        8.5606

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('pd108/pd105')
        1.1849529780564263
        """
        element_abundance = isopy.refval.element.initial_solar_system_abundance_L09
        isotope_fraction = isopy.refval.isotope.initial_solar_system_fraction_L09
        return core.ScalarDict({key: value * element_abundance.get(key.element_symbol) for key, value in isotope_fraction.items()},
                               default_value=np.nan, readonly = True)
        
    @is_reference_value
    @core.cached_property
    def sprocess_fraction_B11(self):
        """
        Dictionary containing the estimated s-process fraction of each isotope from Bisterzo et al. (2011).

        Reference: `Bisterzo et al. (2011) <https://doi.org/10.1111/j.1365-2966.2011.19484.x>`_.

        The ``get()`` method of this dictionary will automatically calculate the ratio of two
        isotopes if both are present the dictionary. The ``get()`` method will return ``np.nan``
        for absent keys.

        Examples
        --------
        >>> isopy.refval.isotope.initial_solar_system_abundance_L09['pd105']
        0.157

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('mo95')
        0.696

        >>> isopy.refval.isotope.initial_solar_system_abundance_L09.get('pd108/pd105')
        4.751592356687898
        """
        return core.ScalarDict(_load_RV_values('isotope_sprocess_fraction_B11', np.float64), default_value=np.nan, readonly = True)


class ReferenceValues:
    """ Reference values useful for working with geochemical data"""
    def __init__(self):
        self.reset()

    def __repr__(self):
        string = 'The following groups of reference values are avaliable:\n\n'
        string += f'mass - {self.__class__.mass.fget.__doc__}\n'
        string += f'element - {self.__class__.element.fget.__doc__}\n'
        string += f'isotope - {self.__class__.isotope.fget.__doc__}'
        return string


    def reset(self):
        """Resets the reference values to the default values"""
        self._element = element()
        self._isotope = isotope()
        self._mass = mass()

    @property
    def isotope(self):
        """Reference values for different isotopes"""
        return self._isotope

    @property
    def element(self):
        """Reference values for different elements"""
        return self._element

    @property
    def mass(self):
        """Reference values for different atomic masses"""
        return self._mass



import os as _os
from isopy import core
import csv as csv
import datetime as dt
import numpy as np
import chardet
import openpyxl
import pyperclip
from openpyxl import load_workbook
import itertools
import io
import numpy as np


__all__ = ['read_exp', 'read_csv', 'write_csv', 'read_xlsx', 'write_xlsx', 'read_clipboard',
           'array_from_csv', 'array_from_xlsx', 'array_from_clipboard']

import isopy.checks


################
### read exp ###
################
class NeptuneData:
    """
    Container for the data returned by ``read_exp``.
    """
    def __init__(self, info, cycle, time, measurements):
        self.info = info
        self.cycle = cycle
        self.time = time
        self.measurements = measurements

def read_exp(filename, rename = None) -> NeptuneData:
    """
    Load data from a Neptune/Triton export file.

    Parameters
    ----------
    filename : str, bytes, StringIO, BytesIO
        Path for file to be opened. Alternatively a file like byte string can be supplied.
        Also accepts file like objects.
    rename : dict, Callable, Optional
        For renaming keys in the analysed data. Useful for cases when the key is the mass rather
        than the isotope measured. If a dictionary is passed then every key present in the
        dictionary will be replaced by the associated value. A callable can also be passed that
        takes one key in the file and returns the new key.


    Returns
    -------
    neptune_data : NeptuneData
        An object containing the following attributes:

        * info - Dictionary containing the metadata included at the beginning of the file.
        * cycle - A list containing the cycle number for each measurement.
        * time - A list containing datetime objects for each measurement.
        * measurements - An dictionary containing an an isopy array with the values in for each line measured. Static measurements are always given as line ``1``.
         to extract e.g only the isotope data from the measurement use ``neptune_data.measurement[line].copy(flavour_eq='isotope')``.
    """
    information = {}

    # If filename is a string load the files.
    if type(filename) is str:
        with open(filename, 'rb') as fileio:
            file = fileio.read()
    elif type(filename) is bytes:
        file = filename
    elif type(filename) is io.BytesIO:
        filename.seek(0)
        file = filename.read()
    elif type(filename) is str:
        file = filename
    elif type(filename) is io.StringIO:
        filename.seek(0)
        file = filename.read()
    else:
        raise TypeError('filename is of unknown type')

    if type(file is bytes):
        # find the files encoding
        encoding = chardet.detect(file).get('encoding')

        # Decode the bytes into string.
        file = file.decode(encoding)

    dialect = csv.Sniffer().sniff(file)

    csv_reader = csv.reader(io.StringIO(file), dialect=dialect)
    for row in csv_reader:
        if ':' in row[0]:  # mete
            name, value = row[0].split(':', 1)
            information[name] = value

        if row[0] == 'Cycle':
            # We can reuse the function from before to read the integration data
            data = _read_csv_ckeys(row, csv_reader, float_prefered=False, termination_symbol='***')

    if rename is None:
        renamer = lambda name: name
    elif isinstance(rename, dict):
        renamer = lambda name: rename.get(name, name)
    elif callable(rename):
        renamer = rename
    else:
        raise TypeError('rename must be a dict or a callable function')

    data.pop("", None)
    cycle = np.array(data.pop('Cycle'), dtype ='int')
    time = data.pop('Time')
    try:
        time =[dt.datetime.strptime(time[i], '%H:%M:%S:%f') for i in range(len(time))]
    except ValueError:
        try:
            time = [dt.datetime.fromtimestamp(time[i]).strftime('%H:%M:%S:%f') for i in range(len(time))]
        except:
            time = [dt.datetime.fromtimestamp(0).strftime('%H:%M:%S:%f') for i in range(len(time))]
    measurements = {}

    for key in data:
        if ":" in key:
            line, keystring = key.split(":", 1)
            line = int(line)
        else:
            line = 1
            keystring = key
        keystring = renamer(keystring)
        measurements.setdefault(line, {})[isopy.keystring(keystring)] = data[key]


    measurements = {line: isopy.asarray(measurements[line]) for line in measurements}
    return NeptuneData(information, cycle, time, measurements)

######################
### read/write CSV ###
######################
def read_csv(filename, comment_symbol ='#', keys_in = 'c',
             float_preferred = False, encoding = None,
             dialect = None):
    """
    Load data from a csv file.



    Parameters
    ----------
    filename : str, bytes, StringIO, BytesIO
        Path for file to be opened. Alternatively a file like byte string can be supplied.
        Also accepts file like objects.
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.
    keys_in : {'c', 'r', None}, Default = 'c'
        If keys are given as the first value in each column pass ``c``. If keys are given as the
        first value in each row pass ``r``. If there are no keys pass ``None``.
    float_preferred : bool, Default = False
        If `True` all values will be converted to float if possible. If conversion fails the
        value will be left as a string.
    encoding
        Encoding of the file. If None the encoding will be guessed from the file.
    dialect
        Dialect of the csv file. If None the dialect will be guessed from the file.

    Returns
    -------
    data : dict or list
        Returns a dictionary for data with keys otherwise a list.
    """
    filename = isopy.checks.check_type('filename', filename, str, bytes)
    comment_symbol = isopy.checks.check_type('comment_symbol', comment_symbol, str)
    keys_in = isopy.checks.check_type('keys_in', keys_in, str,allow_none=True)
    encoding = isopy.checks.check_type('encoding', encoding, str, allow_none=True)
    dialect = isopy.checks.check_type('dialect', dialect, str, allow_none=True)

    #If filename is a string load the files.
    if type(filename) is str:
        with open(filename, 'rb') as fileio:
            file = fileio.read()
    elif type(filename) is bytes:
        file = filename
    elif type(filename) is io.BytesIO:
        filename.seek(0)
        file = filename.read()
    elif type(filename) is str:
        file = filename
    elif type(filename) is io.StringIO:
        filename.seek(0)
        file = filename.read()
    else:
        raise TypeError('filename is of unknown type')

    if type(file is bytes):
        #find the files encoding
        if encoding is None:
            encoding = chardet.detect(file).get('encoding')

        #Decode the bytes into string.
        file = file.decode(encoding)

    #find the csv dialect
    if dialect is None:
        dialect = csv.Sniffer().sniff(file)

    # Create a reader object by converting the file string to a file-like object
    csv_reader = csv.reader(io.StringIO(file), dialect=dialect)
    for row in csv_reader:
        row_data = [r.strip() for r in
                    row]  # Remove any training whitespaces from the data in this row

        if row_data.count('') == len(row):
            # All the columns in this row are empty. Ignore it
            continue

        if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol: #startswith(comment_symbol):
            # This is a comment so ignore it.
            continue

        if keys_in == 'c':
            data =  _read_csv_ckeys(row_data, csv_reader, comment_symbol, float_prefered=float_preferred)
            data.pop("", None)
            return data
        elif keys_in == 'r':
            data =  _read_csv_rkeys(row_data, csv_reader, comment_symbol, float_prefered=float_preferred)
            data.pop("", None)
            return data
        elif keys_in is None:
            return _read_csv_nokeys(row_data, csv_reader, comment_symbol, float_prefered=float_preferred)
        else:
            raise ValueError(f'Unknown value for "keys_in" {keys_in}')

def read_clipboard(comment_symbol ='#', keys_in = 'c',
             float_preferred = False, dialect = None):
    """
    Load data from values in the clipboard.

    Parameters
    ----------
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.
    keys_in : {'c', 'r', None}, Default = 'c'
        If keys are given as the first value in each column pass ``c``. If keys are given as the
        first value in each row pass ``r``. If there are no keys pass ``None``.
    float_preferred : bool, Default = False
        If `True` all values will be converted to float if possible. If conversion fails the
        value will be left as a string.
    dialect
        Dialect of the values in the clipboard. If None the dialect will be guessed from the values.

    Returns
    -------
    data : dict or list
        Returns a dictionary for data with keys otherwise a list.
    """
    data = pyperclip.paste()
    data = data.encode('UTF-8')
    return read_csv(data, encoding='UTF-8', comment_symbol=comment_symbol, keys_in=keys_in,
                    float_preferred=float_preferred, dialect=dialect)


def _read_csv_ckeys(first_row, reader, comment_symbol=None, termination_symbol=None, float_prefered = False):
    keys = first_row  # First row contains the the name of each column
    values = [[] for i in range(len(keys))]  # Create an empty list for each key

    # Loop over the remaining rows
    for i, row in enumerate(reader):
        row = [r.strip() for r in row]

        if termination_symbol is not None and row[0][:len(termination_symbol)] == termination_symbol: #startswith(termination_symbol):
            # Stop reading data if we find this string at the beginning of a row
            break

        if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol: #.startswith(comment_symbol):
            # Row is a comment, ignore
            continue

        if row.count('') == len(row):
            # Empty row, ignore
            continue

        if len(row) < len(keys):
            # There is not enough values in this row to give one to each key
            raise ValueError(f'Row {i} does not have enough columns')

        for j in range(len(keys)):
            value = row[j]

            if float_prefered:
                try:
                    value = float(value)
                except:
                    value = str(value)

            values[j].append(value)

    return dict(zip(keys, values))  # creates a dictionary from two lists


def _read_csv_rkeys(first_row, reader, comment_symbol=None, termination_symbol=None,
                    float_prefered=False):
    data = {}
    dlen = None

    # Loop over the remaining rows
    for i, row in enumerate(itertools.chain([first_row], reader)):
        row = [r.strip() for r in row]

        if termination_symbol is not None and row[0][:len(termination_symbol)] == termination_symbol: #.startswith(termination_symbol):
            # Stop reading data if we find this string at the beginning of a row
            break

        if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol: #.startswith(comment_symbol):
            # Row is a comment, ignore
            continue

        if row.count('') == len(row):
            # Empty row, ignore
            continue

        data[row[0]] = []
        if dlen is None: dlen = len(row)
        elif dlen != len(row):
            raise ValueError('Rows have different number of values')

        for j in range(1, len(row)):
            value = row[j]

            if float_prefered:
                try:
                    value = float(value)
                except:
                    value = str(value)

            data[row[0]].append(value)

    return data


def _read_csv_nokeys(first_row, reader, comment_symbol=None, termination_symbol=None,
                    float_prefered=False):
    data = []
    dlen = None

    # Loop over the remaining rows
    for i, row in enumerate(itertools.chain([first_row], reader)):
        row = [r.strip() for r in row]

        if termination_symbol is not None and row[0][:len(termination_symbol)] == termination_symbol: #.startswith(termination_symbol):
            # Stop reading data if we find this string at the beginning of a row
            break

        if comment_symbol is not None and row[0][:len(comment_symbol)] == comment_symbol: #.startswith(comment_symbol):
            # Row is a comment, ignore
            continue

        if row.count('') == len(row):
            # Empty row, ignore
            continue

        data.append([])
        if dlen is None:
            dlen = len(row)
        elif dlen != len(row):
            raise ValueError('Rows have different number of values')

        for j in range(len(row)):
            value = row[j]

            if float_prefered:
                try:
                    value = float(value)
                except:
                    value = str(value)

            data[-1].append(value)

    return data


def write_csv(filename, data, comments=None, keys_in='c',
              delimiter = ',', comment_symbol = '#' ) -> None:
    """
    Save data to a csv file.

    Parameters
    ----------
    filename : str, StringIO, BytesIO
        Path/name of the csv file to be created. Any existing file with the same path/name will be
        over written. Also accepts file like objects.
    data : isopy_array_like, numpy_array_like
        Data to be saved in the array.
    comments : str, Sequence[str], Optional
        Comments to be included at the top of the file
    keys_in : {'c', 'r'}, Default = 'c'
        Only applicable for isopy arrays like *data*. For keys as the first value in each column
        pass ``"c"``. For keys as the first value in each row pass ``"r"``.
    delimiter : str, Default = ','
        This string will be used to separate values in the file
    comment_symbol : str, Default = '#'
        This string will precede any comments at the beginning of the file.
    """
    #filename = isopy.checks.check_type('filename', filename, str)
    comments = isopy.checks.check_type_list('comments', comments, str, allow_none=True, make_list=True)
    keys_in = isopy.checks.check_type('keys_in', keys_in, str, allow_none=True)
    delimiter = isopy.checks.check_type('delimiter', delimiter, str)
    comment_symbol = isopy.checks.check_type('comment_symbol', comment_symbol, str)

    data = isopy.asanyarray(data)
    if isinstance(data, core.IsopyArray):
        if data.ndim == 0:
            data = data.reshape(-1)

        dsize = data.nrows
        csize = data.ncols
        ndim = data.ndim
        data = data.to_dict()

        if keys_in == 'c':
            func = _write_csv_ckeys
            csize = csize - 1
        elif keys_in == 'r':
            func = _write_csv_rkeys
            csize = dsize
        elif keys_in is None:
            func = _write_csv_nokeys
            data = np.array(data.to_array())
            csize = csize - 1
    else:
        ndim = data.ndim
        if ndim == 0:
            data = data.reshape((1,1))
        elif ndim == 1:
            data = data.reshape((1, -1))
        elif ndim > 2:
            raise ValueError('Cannot save data with more than two dimensions')
        dsize = data.shape[0]
        csize = data.shape[1] - 1
        func = _write_csv_nokeys

    rows = []

    if comments is not None:
        for comment in comments:
            rows.append([f'{comment_symbol}{comment}'] + ['' for i in range(csize)])

    rows += func(data, dsize)
    if type(filename) is str:
        with open(filename, mode='w', newline='') as file:
            csv_writer = csv.writer(file, delimiter=delimiter)
            for row in rows:
                csv_writer.writerow(row)

    if type(filename) is io.StringIO:
        csv_writer = csv.writer(filename, delimiter=delimiter)
        for row in rows:
            csv_writer.writerow(row)

    if type(filename) is io.BytesIO:
        file = io.StringIO()
        csv_writer = csv.writer(file, delimiter=delimiter)
        for row in rows:
            csv_writer.writerow(row)
        file.seek(0)
        string = file.read()
        filename.write(string.encode('UTF-8'))

def _write_csv_ckeys(data, dsize):
    keys = data.keys()
    rows = []
    rows.append([f'{key}' for key in keys])

    for i in range(dsize):
        rows.append([f'{data[key][i]}' for key in keys])

    return rows

def _write_csv_rkeys(data, dsize):
    keys = data.keys()
    rows = []

    for key in keys:
        rows.append([f'{key}'] + [f'{data[key][i]}' for i in range(dsize)])
    return rows

def _write_csv_nokeys(data, dsize):
    rows = []
    for i in range(data.shape[0]):
        rows.append([f'{data[i][j]}' for j in range(data.shape[-1])])
    return rows

#######################
### Read/write xlsx ###
#######################
def read_xlsx(filename, sheetname=None, keys_in='c',
              default_value = np.nan, comment_symbol='#'):
    """
    Load data from an excel file.

    Parameters
    ----------
    filename : str, bytes, BytesIO
        Path for file to be opened. Also accepts file like objects.
    sheetname : str, int, Optional
        To load data from a single sheet in the workbook pass either the name of the sheet or
        the position of the sheet. If nothing is specified all the data for all sheets in the
        workbook is returned.
    keys_in : {'c', 'r', None}, Default = 'c'
        If keys are given as the first value in each column pass ``c``. If keys are given as the
        first value in each row pass ``r``. If there are no keys pass ``None``.
    default_value : scalar, Default = np.nan
        Value substituted for empty cells or cells containing errors
    comment_symbol : str, Default = '#'
        Rows starting with this string will be ignored.

    Returns
    -------
    data : dict
        A dictionary containing the data for a single sheet or a dictionary containing the data
        for multiple sheets in the workbook indexed by sheet name.
    """
    if filename is bytes:
        filename = io.BytesIO(filename)

    workbook = openpyxl.load_workbook(filename, data_only=True, read_only=True)

    sheetnames = workbook.sheetnames

    if sheetname is None:
        sheet_data = {}
        for sheetname in sheetnames:
            sheet_data[sheetname] = _read_xlsx_sheet(workbook[sheetname], keys_in, comment_symbol, default_value)
        return sheet_data

    elif type(sheetname) is int:
        return _read_xlsx_sheet(workbook[sheetnames[sheetname]], keys_in, comment_symbol, default_value)

    elif sheetname in sheetnames:
        return _read_xlsx_sheet(workbook[sheetname], keys_in, comment_symbol, default_value)

    else:
        raise ValueError(f'sheetname "{sheetname}" not found in workbook"')


def _read_xlsx_sheet(worksheet, keys_in, comment_symbol, default_value):
    for ri in range(1, worksheet.max_row + 1):
        cell = worksheet.cell(ri, 1)
        if cell.data_type == 's' and cell.value[:len(comment_symbol)] == comment_symbol:
            continue

        if keys_in == 'c':
            return _read_xlsx_sheet_ckeys(worksheet, ri, default_value)
        elif keys_in == 'r':
            return _read_xlsx_sheet_rkeys(worksheet, ri, default_value)
        elif keys_in is None:
            return _read_xlsx_sheet_nokeys(worksheet, ri, default_value)
        else:
            raise ValueError(f'Unknown value for "keys_in" {keys_in}')

def _read_xlsx_sheet_ckeys(worksheet, row_index, default_value):
    data = {}
    ERROR = object()

    # Iterate over each column
    for ci in range(1, worksheet.max_column + 1):
        # First value is the key
        key = f'{worksheet.cell(row_index, ci).value}'.strip()

        if key == '':
            continue

        data[key] = []

        for ri in range(row_index + 1, worksheet.max_row + 1):
            cell = worksheet.cell(ri, ci)
            value = cell.value
            if type(value) is str:
                value = value.strip()
            if cell.data_type == 'e' or (type(value) is str and value.upper()) == '=NA()':
                data[key].append(ERROR)
            elif cell.data_type == 's':
                if value == '':
                    data[key].append(None)
                else:
                    data[key].append(value)
            else:
                data[key].append(cell.value)

    keys = list(data.keys())
    for i in range(len(data[keys[0]])):
        is_none = [(data[key][i] is None) for key in keys]
        if is_none.count(True) == len(keys):
            for key in keys:
                data[key].pop(i)
        else:
            for j, key in enumerate(keys):
                if is_none[j] or data[key][i] is ERROR:
                    data[key][i] = default_value
    return data

def _read_xlsx_sheet_rkeys(worksheet, row_index, default_value):
    data = {}
    ERROR = object()

    # Iterate over each column
    for ri in range(row_index, worksheet.max_row + 1):
        # First value is the key
        key = f'{worksheet.cell(ri, 1).value}'.strip()

        if key == '':
            continue

        data[key] = []

        for ci in range(2, worksheet.max_column + 1):
            cell = worksheet.cell(ri, ci)
            value = cell.value
            if type(value) is str:
                value = value.strip()
            if cell.data_type == 'e' or (type(value) is str and value.upper()) == '=NA()':
                data[key].append(ERROR)
            elif cell.data_type == 's':
                if value == '':
                    data[key].append(None)
                else:
                    data[key].append(value)
            else:
                data[key].append(cell.value)

    keys = list(data.keys())
    for i in range(len(data[keys[0]])):
        is_none = [(data[key][i] is None) for key in keys]
        if is_none.count(True) == len(keys):
            for key in keys:
                data[key].pop(i)
        else:
            for j, key in enumerate(keys):
                if is_none[j] or data[key][i] is ERROR:
                    data[key][i] = default_value
    return data

def _read_xlsx_sheet_nokeys(worksheet, row_index, default_value):
    data = []
    ERROR = object()

    # Iterate over each column
    for ri in range(row_index, worksheet.max_row + 1):
        # First value is the key
        data.append([])

        for ci in range(1, worksheet.max_column + 1):
            cell = worksheet.cell(ri, ci)
            value = cell.value
            if type(value) is str:
                value = value.strip()
            if cell.data_type == 'e' or (type(value) is str and value.upper()) == '=NA()':
                data[-1].append(ERROR)
            elif cell.data_type == 's':
                if value == '':
                    data[-1].append(None)
                else:
                    data[-1].append(value)
            else:
                data[-1].append(cell.value)

        if data[-1].count(None) == len(data[-1]):
            data.pop(-1)
        else:
            for i, v in enumerate(data[-1]):
                if v is None or v is ERROR:
                    data[-1][i] = default_value
    return data

#TODO change np.nan values to '=NA()'
def write_xlsx(filename, *sheets, comments = None, comment_symbol= '#',
               keys_in= 'c', append = False, clear = True, first_cell ="A1", **sheetnames):
    """
    Save data to an excel file.

    Parameters
    ----------
    filename : str, BytesIO
        Path/name of the excel file to be created. Any existing file with the same path/name
        will be overwritten. Also accepts file like objects.
    sheets : isopy_array_like, numpy_array_like
        Data arrays given here will be saved as sheet1, sheet2 etc.
    comments : str, Sequence[str], Optional
        Comments to be included at the top of the file
    comment_symbol : str, Default = '#'
        This string will precede any comments at the beginning of the file
    keys_in : {'c', 'r'}, Default = 'c'
        Only applicable for isopy arrays like *data*. For keys as the first value in each column
        pass ``"c"``. For keys as the first value in each row pass ``"r"``.
    append : bool, Default = False
        If ``True`` and *filename* exists it will append the data to this workbook. An exception
        is raised if *filename* is not a valid excel workbook.
    clear : bool, Default = True
        If ``True`` any preexisting sheets are cleared before any new data is written to it.
    first_cell: str
        The first cell where the data is written. Defaults value is ``"A1"``.
    sheetnames : isopy_array, numpy_array
        Data array given here will be saved in the workbook with the keyword as the sheet name
    """
    save = True
    if type(filename) is openpyxl.Workbook:
        workbook = filename
        save = False
    elif type(filename) is io.BytesIO:
        if append:
            #TODO some sort of check to make sure the file isnt empty
            try:
                workbook = openpyxl.load_workbook(filename=filename)
            except:
                workbook = openpyxl.Workbook()
                workbook.remove(workbook.active)
        else:
            workbook = openpyxl.Workbook()
            workbook.remove(workbook.active)
    elif type(filename) is str:
        if append and _os.path.exists(filename):
            workbook = openpyxl.load_workbook(filename=filename)
        else:
            workbook = openpyxl.Workbook()
            workbook.remove(workbook.active)

    sheetname_data = {f'sheet{i+1}': data for i, data in enumerate(sheets)}
    sheetname_data.update(sheetnames)
    try:
        for sheetname, data in sheetname_data.items():
            #If appending then delete any preexisting workbooks
            if sheetname not in workbook.sheetnames:
                worksheet = workbook.create_sheet(sheetname)
            elif clear:
                workbook.remove(workbook[sheetname])
                worksheet = workbook.create_sheet(sheetname)
            else:
               worksheet = workbook[sheetname]

            _write_xlsx(worksheet, data, comments, comment_symbol, keys_in, first_cell)

        #Workbooks must have at least one sheet
        if len(workbook.sheetnames) == 0:
            workbook.create_sheet('sheet1')

        if save: workbook.save(filename)
    finally:
        if save: workbook.close()

def _write_xlsx(worksheet, data, comments, comment_symbol, keys_in, cell):
    data = isopy.asanyarray(data)

    if type(cell) is tuple:
        ri, ci = cell
    else:
        ri, ci = openpyxl.utils.cell.coordinate_to_tuple(cell)

    if comments is not None:
        if type(comments) is not list:
            comments = [comments]
        for comment in comments:
            worksheet.cell(ri, ci).value = f'{comment_symbol}{comment}'
            ri += 1

    if isinstance(data, isopy.core.IsopyArray):
        if data.ndim == 0:
            data = data.reshape(-1)
        if keys_in == 'c':
            data = data.to_dict()
            _write_xlsx_ckeys(worksheet, data, ri, ci)
        elif keys_in == 'r':
            data = data.to_dict()
            _write_xlsx_rkeys(worksheet, data, ri, ci)
        elif keys_in is None:
            data = np.array(data.to_list())
            _write_xlsx_nokeys(worksheet, data, ri, ci)
    else:
        ndim = data.ndim
        if ndim == 0:
            data = data.reshape((1, 1))
        elif ndim == 1:
            data = data.reshape((1, -1))
        elif ndim > 2:
            raise ValueError('Cannot save data with more than two dimensions')
        _write_xlsx_nokeys(worksheet, data, ri, ci)


def _write_xlsx_ckeys(worksheet, data, ri, ci):
    for cj, key in enumerate(data.keys()):
        worksheet.cell(ri, ci + cj).value = f'{key}'

        for rj, value in enumerate(data[key]):
            if np.isnan(value):
                value = '#N/A'
            worksheet.cell(ri + rj + 1, ci + cj).value = value

def _write_xlsx_rkeys(worksheet, data, ri, ci):
    for rj, key in enumerate(data.keys()):
        worksheet.cell(ri + rj, ci).value = f'{key}'

        for cj, value in enumerate(data[key]):
            if np.isnan(value):
                value = '#N/A'
            worksheet.cell(ri + rj, ci + cj + 1).value = value

def _write_xlsx_nokeys(worksheet, data, ri, ci):
    for rj, row in enumerate(data):
        for cj, value in enumerate(row):
            if np.isnan(value):
                value = '#N/A'
            worksheet.cell(ri + rj, ci+ cj).value = value

################
### to array ###
################

def array_from_csv(filename, *, dtype=None, ndim=None, **read_csv_kwargs):
    """
    Returns an array of values from a csv file.
    """
    data = read_csv(filename, **read_csv_kwargs)
    return isopy.asanyarray(data, dtype=dtype, ndim=ndim)

def array_from_xlsx(filename, sheetname, *, dtype=None, ndim=None, **read_xlsx_kwargs):
    """
    Returns an array of values from *sheet_name* in a xlsx file.
    """
    data = read_xlsx(filename, sheetname, **read_xlsx_kwargs)
    return isopy.asanyarray(data, dtype=dtype, ndim=ndim)

def array_from_clipboard(*, dtype=None, ndim=None, **read_clipboard_kwargs):
    """
    Returns an array of values from the clipboard.
    """
    data = read_clipboard(**read_clipboard_kwargs)
    return isopy.asanyarray(data, dtype=dtype, ndim=ndim)


import numpy as np
from numpy.lib.function_base import array_function_dispatch
from scipy import stats
import functools
from . import core

__all__ = ['sd', 'nansd', 'se', 'nanse', 'mad', 'nanmad',
           'nancount',
           'add', 'subtract', 'divide', 'multiply', 'power', 'arrayfunc',
           'keymax', 'keymin']
__all__ += 'sd2 sd3 sd95 sd99 nansd2 nansd3 nansd95 nansd99'.split()
__all__ += 'se2 se3 se95 se99 nanse2 nanse3 nanse95 nanse99'.split()
__all__ += 'mad2 mad3 mad95 mad99 nanmad2 nanmad3 nanmad95 nanmad99'.split()

# TODO homoscedastic_sd(*args)

def calculate_ci(ci, df=None):
    if df is None:
        return stats.norm.ppf(0.5 + ci / 2)
    else:
        return stats.t.ppf(0.5 + ci / 2, df)

##########################
### Dispatch functions ###
##########################
def _sd_dispatcher(a, axis=None, *, ci = None, zscore=None): #, where=None):
    return (a,)

@core.set_module('isopy')
@array_function_dispatch(_sd_dispatcher)
def sd(a, axis = None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom.

    Shortcut for ``np.std(a, ddof = 1, axis=axis)``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``sd2``, ``sd3``, ``sd95`` ``sd99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.sd(array)
    (row) , Ru  , Pd      , Cd
    None  , nan , 0.60553 , 1.41745
    >>> isopy.sd(array, axis=1)
    array([       nan, 2.35867194, 1.28970281, 3.17962262])
    >>> isopy.sd2(array) #same as sd(array, zscore=2)
    (row) , Ru  , Pd      , Cd
    None  , nan , 1.21106 , 2.83490
    >>> isopy.sd95(array) #same as sd(array, ci=0.95)
    (row) , Ru  , Pd      , Cd
    None  , nan , 1.9271 , 4.511

    See Also
    --------
    :func:`nansd`, `np.std <https://numpy.org/devdocs/reference/generated/numpy.std.html#numpy.std>`_
    """
    result = np.std(a, ddof=1, axis=axis)
    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

#TODO
def homoscedastic_sd(*a, axis=1):
    raise NotImplementedError()
    ddof = len(a)
    a = np.concatenate(a) #wont work if a is not 1 dimensional
    return np.std(a, ddof = ddof)


@core.set_module('isopy')
@array_function_dispatch(_sd_dispatcher)
def nansd(a, axis = None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard deviation along the specified axis for N-1 degrees of freedom, while ignoring NaNs.

    Shortcut for ``np.nanstd(a, ddof = 1, axis=axis)``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nansd2``, ``nansd3``, ``nansd95`` ``nansd99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.


    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nansd(array)
    (row) , Ru      , Pd      , Cd
    None  , 0.55678 , 0.60553 , 1.41745
    >>> isopy.nansd(array, axis=1)
    array([2.12132034, 2.35867194, 1.28970281, 3.17962262])
    >>> isopy.nansd2(array) #same as nansd(array, zscore=2)
    (row) , Ru      , Pd      , Cd
    None  , 1.11355 , 1.21106 , 2.83490
    >>> isopy.nansd95(array) #same as nansd(array, ci=0.95)
    (row) , Ru      , Pd      , Cd
    None  , 2.3956 , 1.9271 , 4.511

    See Also
    --------
    :func:`sd`, `numpy.nanstd <https://numpy.org/devdocs/reference/generated/numpy.nanstd.html#numpy.nanstd>`_
    """
    result = np.nanstd(a, ddof=1, axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

def _se_dispatcher(a, axis=None, *, ci = None, zscore=None): #, where = None):
    return (a,)

@core.set_module('isopy')
@array_function_dispatch(_se_dispatcher)
def se(a, axis=None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom.

    Shortcut for ``scipy.stats.sem(a, axis=axis, nan_policy='propagate')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``se2``, ``se3``, ``se95`` ``se99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.se(array)
    (row) , Ru  , Pd      , Cd
    None  , nan , 0.30277 , 0.70873
    >>> isopy.se(array, axis=1)
    array([       nan, 1.36177988, 0.74461026, 1.83575598])
    >>> isopy.se2(array) #same as se(array, zscore=2)
    (row) , Ru  , Pd      , Cd
    None  , nan , 0.60553 , 1.41745
    >>> isopy.se95(array) #same as se(array, ci=0.95)
    (row) , Ru  , Pd      , Cd
    None  , nan , 0.96353 , 2.2555

    See Also
    --------
    :func:`nanse`, `scipy.stats.sem <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.sem.html?highlight=sem#scipy.stats.sem>`_
    """
    result =  stats.sem(a, nan_policy='propagate', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

@core.set_module('isopy')
@array_function_dispatch(_se_dispatcher)
def nanse(a, axis=None, *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the standard error along the specified axis for N-1 degrees of freedom, while ignoring
    NaNs.

    Shortcut for ``scipy.stats.sem(a, axis=axis, nan_policy='omit')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nanse2``, ``nanse3``, ``nanse5`` ``nanse99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nanse(array)
    (row) , Ru      , Pd      , Cd
    None  , 0.32146 , 0.30277 , 0.70873
    >>> isopy.nanse(array, axis=1) #returns a masked array
    masked_array(data=[1.4999999999999996, 1.3617798810543664,
                   0.7446102634562892, 1.835755975068582],
             mask=[False, False, False, False],
       fill_value=1e+20)
    >>> isopy.nanse2(array) #same as nanse(array, zscore=2)
    (row) , Ru      , Pd      , Cd
    None  , 0.64291 , 0.60553 , 1.41745
    >>> isopy.nanse95(array) #same as nanse(array, ci=0.95)
    (row) , Ru      , Pd      , Cd
    None  , 1.3831 , 0.96353 , 2.2555

    See Also
    --------
    :func:`se`, `scipy.stats.sem <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.sem.html?highlight=sem#scipy.stats.sem>`_
    """
    result = stats.sem(a, nan_policy='omit', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

def _mad_dispatcher(a, axis=None, scale=None, *, ci = None, zscore=None): #, where = None):
    return (a,)

@core.set_module('isopy')
@array_function_dispatch(_mad_dispatcher)
def mad(a, axis=None, scale= 'normal', *, ci = None, zscore=None): #,  where = core.NotGiven):
    """
    Compute the median absolute deviation of the data along the given axis.

    Shortcut for ``scipy.stats.median_abs_deviation(a, axis, scale=scale, nan_policy='propagate')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``mad2``, ``mad3``, ``mad95`` ``mad99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.mad(array)
    (row) , Ru  , Pd      , Cd
    None  , nan , 0.66717 , 1.03782
    >>> isopy.mad(array, axis=1)
    array([       nan, 2.96520444, 1.03782155, 3.55824532])
    >>> isopy.mad2(array) #same as mad(array, zscore=2)
    (row) , Ru  , Pd      , Cd
    None  , nan , 1.33434 , 2.07564
    >>> isopy.mad95(array) #same as mad(array, ci=0.95)
    (row) , Ru  , Pd      , Cd
    None  , nan , 2.1232 , 3.3028

    See Also
    --------
    :func:`nanmad`, `scipy.stats.median_abs_deviation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html#scipy.stats.median_abs_deviation>`_
    """
    result = stats.median_abs_deviation(a, scale=scale, nan_policy='propagate', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

@core.set_module('isopy')
@array_function_dispatch(_mad_dispatcher)
def nanmad(a, axis=None, scale = 'normal', *, ci = None, zscore=None): #, where = core.NotGiven):
    """
    Compute the median absolute deviation along the specified axis, while ignoring NaNs.

    Shortcut for ``scipy.stats.median_abs_deviation(a, axis, scale=scale, nan_policy='omit')``.

    If *ci* is given the result will be multiplied by the confidence interval *ci* of a
    t-distribution with N-1 degrees of freedom.

    If *zscore* is given then the result is multiplied by *zscore*.

    Functions ``nanmad2``, ``nanmad3``, ``nanmad95`` ``nanmad99`` are also avaliable for z-score
    values of 2 and 3 and for 95% and 99% confidence intervals respectivley.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nanmad(array)
    (row) , Ru      , Pd      , Cd
    None  , 0.59304 , 0.66717 , 1.03782
    >>> isopy.nanmad(array, axis=1)
    array([2.22390333, 2.96520444, 1.03782155, 3.55824532])
    >>> isopy.nanmad2(array) #same as nanmad(array, zscore=2)
    (row) , Ru      , Pd      , Cd
    None  , 1.18608 , 1.33434 , 2.07564
    >>> isopy.nanmad95(array) #same as nanmad(array, ci=0.95)
    (row) , Ru      , Pd      , Cd
    None  , 2.5516 , 2.1232 , 3.3028

    See Also
    --------
    :func:`mad`, `scipy.stats.median_abs_deviation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.median_abs_deviation.html#scipy.stats.median_abs_deviation>`_
    """
    result = stats.median_abs_deviation(a, scale=scale, nan_policy = 'omit', axis=axis)

    if ci is not None:
        df = np.count_nonzero(np.invert(np.isnan(a)), axis=axis) - 1
        return result * calculate_ci(ci, df)

    if zscore is not None:
        return result * zscore

    return result

### preset functions

sd2 = core.partial_func(sd, 'sd2', zscore=2)
sd3 = core.partial_func(sd, 'sd3', zscore=3)
sd95 = core.partial_func(sd, 'sd95', ci=0.95)
sd99 = core.partial_func(sd, 'sd99', ci=0.99)
nansd2 = core.partial_func(nansd, 'nansd2', zscore=2)
nansd3 = core.partial_func(nansd, 'nansd3', zscore=3)
nansd95 = core.partial_func(nansd, 'nansd95', ci=0.95)
nansd99 = core.partial_func(nansd, 'nansd99', ci=0.99)

se2 = core.partial_func(se, 'se2', zscore=2)
se3 = core.partial_func(se, 'se3', zscore=3)
se95 = core.partial_func(se, 'se95', ci=0.95)
se99 = core.partial_func(se, 'se99', ci=0.99)
nanse2 = core.partial_func(nanse, 'nanse2', zscore=2)
nanse3 = core.partial_func(nanse, 'nanse3', zscore=3)
nanse95 = core.partial_func(nanse, 'nanse95', ci=0.95)
nanse99 = core.partial_func(nanse, 'nanse99', ci=0.99)

mad2 = core.partial_func(mad, 'mad2', zscore=2)
mad3 = core.partial_func(mad, 'mad3', zscore=3)
mad95 = core.partial_func(mad, 'mad95', ci=0.95)
mad99 = core.partial_func(mad, 'mad99', ci=0.99)
nanmad2 = core.partial_func(nanmad, 'nanmad2', zscore=2)
nanmad3 = core.partial_func(nanmad, 'nanmad3', zscore=3)
nanmad95 = core.partial_func(nanmad, 'nanmad95', ci=0.95)
nanmad99 = core.partial_func(nanmad, 'nanmad99', ci=0.99)

def _count_dispatcher(a, axis=None): #, where = None):
    return (a,)


@core.set_module('isopy')
@array_function_dispatch(_count_dispatcher)
def nancount(a, axis=None): #, where = core.NotGiven):
    """
    Count all values in array that are not NaN or Inf along the specified axis.

    Shortcut for ``np.count_nonzero(np.isfinite(a), axis=axis)``.

    Examples
    --------
    >>> array = isopy.array(ru = [np.nan, 1.1, 2.2, 1.8],
                            pd = [3.1, 3.8, 2.9, 4.2],
                            cd = [6.1, 5.8, 4.7, 8.1])
    >>> isopy.nancount(array)
    (row) , Ru      , Pd      , Cd
    None  , 3.00000 , 4.00000 , 4.00000
    >>> isopy.nancount(array, axis=1)
    array([2, 3, 3, 3], dtype=int64)
    """
    #if where is not core.NotGiven:
    #    a = a[where]

    return np.count_nonzero(np.isfinite(a), axis=axis)

#####################################
### Functions without dispatchers ###
#####################################
# These should  only be used with isopy arrays
@core.set_module('isopy')
def add(x1, x2, default_value=np.nan, keys=None):
    """
    Add *x2* to *x1* substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a default value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    Examples
    --------
    >>> array1 = isopy.ones(1, ('ru', 'pd'))
    >>> array2 = isopy.ones(1, ('pd', 'cd'))
    >>> isopy.add(array1, array2) #Same as array1 + array2
    (row) , Ru  , Pd      , Cd
    0     , nan , 2.00000 , nan
    >>> isopy.add(array1, array2, 0)
    (row) , Ru      , Pd      , Cd
    0     , 1.00000 , 2.00000 , 1.00000
    """
    return core.call_array_function(np.add, x1, x2, default_value=default_value, keys=keys)

@core.set_module('isopy')
def subtract(x1, x2, default_value=np.nan, keys=None):
    """
    Subtract *x2* from *x1* substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a default value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    Examples
    --------
    >>> array1 = isopy.ones(1, ('ru', 'pd'))
    >>> array2 = isopy.ones(1, ('pd', 'cd'))
    >>> isopy.subtract(array1, array2) #Same as array1 - array2
    (row) , Ru  , Pd      , Cd
    0     , nan , 0.00000 , nan
    >>> isopy.subtract(array1, array2, 0)
    (row) , Ru      , Pd      , Cd
    0     , 1.00000 , 0.00000 , -1.00000
    """
    return core.call_array_function(np.subtract, x1, x2, default_value=default_value, keys=keys)

@core.set_module('isopy')
def multiply(x1, x2, default_value=np.nan, keys=None):
    """
    Multiply *x1* by *x2* substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a default value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    Examples
    --------
    >>> array1 = isopy.ones(1, ('ru', 'pd')) * 2
    >>> array2 = isopy.ones(1, ('pd', 'cd')) * 5
    >>> isopy.multiply(array1, array2) #Same as array1 * array2
    (row) , Ru  , Pd       , Cd
    0     , nan , 10.00000 , nan
    >>> isopy.multiply(array1, array2, 1)
    (row) , Ru      , Pd       , Cd
    0     , 2.00000 , 10.00000 , 5.00000
    """
    return core.call_array_function(np.multiply, x1, x2, default_value=default_value, keys=keys)

@core.set_module('isopy')
def divide(x1, x2, default_value=np.nan, keys=None):
    """
    Divide *x1* by *x2* substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a default value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    Examples
    --------
    >>> array1 = isopy.ones(1, ('ru', 'pd')) * 2
    >>> array2 = isopy.ones(1, ('pd', 'cd')) * 5
    >>> isopy.divide(array1, array2) #Same as array1 / array2
    (row) , Ru  , Pd      , Cd
    0     , nan , 0.40000 , nan
    >>> isopy.divide(array1, array2, 1)
    (row) , Ru      , Pd      , Cd
    0     , 2.00000 , 0.40000 , 0.20000
    """
    return core.call_array_function(np.divide, x1, x2, default_value=default_value, keys=keys)

@core.set_module('isopy')
def power(x1, x2, default_value=np.nan, keys=None):
    """
    Raise *x1* to the power of *x2* substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    Examples
    --------
    >>> array1 = isopy.ones(1, ('ru', 'pd')) * 2
    >>> array2 = isopy.ones(1, ('pd', 'cd')) * 5
    >>> isopy.power(array1, array2) #Same as array1 ** array2
    (row) , Ru  , Pd       , Cd
    0     , nan , 32.00000 , nan
    >>> isopy.power(array1, array2, 1)
    (row) , Ru      , Pd       , Cd
    0     , 2.00000 , 32.00000 , 1.00000
    """
    return core.call_array_function(np.power, x1, x2, default_value=default_value, keys=keys)

@core.set_module('isopy')
def arrayfunc(func, *inputs, default_value=np.nan, keys=None, **kwargs):
    """
    Call a numpy function *func* on the *inputs* values substituting *default_value* for absent columns.

    *default_values* can either be a single value which will be used for
    all the inputs or a tuple containing a value for each input.
    If *keys* are given then the operation if only performed for those
    keys.

    This function is useful for calling numpy functions with the isopy specific
    *default_value* and *keys* arguments or for array functions that dont employ
    numpys array function dispatch.

    Examples
    --------
    >>> array = isopy.random(100, keys=('ru', 'pd', 'cd')
    >>> isopy.arrayfunc(np.std, array, keys=('ru', 'cd'))
    (row) , Ru      , Cd
    None  , 0.90836 , 0.89753
    >>> isopy.arrayfunc(scipy.stats.sem, array) #Scipy functions do not support isopy arrays natively
    (row) , Ru      , Pd      , Cd
    None  , 0.09129 , 0.09430 , 0.09021
    """
    return core.call_array_function(func, *inputs, default_value=default_value, keys=keys, **kwargs)

@core.set_module('isopy')
@core.append_preset_docstring
@core.add_preset('abs', abs = True)
def keymax(a, evalfunc = np.nanmedian, abs=False):
    """
    Return the name of the column where the largest value returned by *evalfunc* is found.

    If *abs* is ``True`` the absolute value is used.

    This function requires that *a* is an isopy array.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd')
    >>> isopy.keymax(array)
    IsotopeKeyString('106Pd')
    """
    array = core.asarray(a)
    if abs:
        key_index = np.nanargmax([evalfunc(np.abs(v)) for v in array.values()])
    else:
        key_index = np.nanargmax([evalfunc(v) for v in array.values()])
    return array.keys[key_index]

@core.set_module('isopy')
@core.append_preset_docstring
@core.add_preset('abs', abs = True)
def keymin(a, evalfunc= np.nanmedian, abs = False):
    """
    Return the name of the column where the smallest value returned by *evalfunc* is found.

    If *abs* is ``True`` the absolute value is used.

    This function requires that *a* is an isopy array.

    Examples
    --------
    >>> array = isopy.tb.make_ms_array('pd')
    >>> isopy.keymin(array)
    IsotopeKeyString('102Pd')
    """
    array = core.asarray(a)
    if abs:
        key_index = np.nanargmin([evalfunc(np.abs(v)) for v in array.values()])
    else:
        key_index = np.nanargmin([evalfunc(v) for v in array.values()])
    return array.keys[key_index]

